#include <linux/version.h>
#include <linux/pci.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mutex.h>
#include <linux/videodev2.h>
#include <linux/kthread.h>
#include <linux/freezer.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>

#include "Drivers/XI006AEPRO_Device.h"
#include "v4l2.h"
#include "alsa.h"

unsigned int xi_debug = 0;
module_param(xi_debug,int,0644);
MODULE_PARM_DESC(xi_debug,"enable debug messages");

MODULE_AUTHOR("zhh qianceeli@yahoo.com.cn");
MODULE_DESCRIPTION("xi capture");
MODULE_LICENSE("Dual BSD/GPL");
MODULE_SUPPORTED_DEVICE("{{XI,XI006AEDevice}}");

static struct pci_device_id xi006_pci_tbl[] = 
{
	{ PCI_DEVICE(XI_006AEPRO_VENDOR_ID,XI_006AEPRO_DEVICE_ID) },
	{ 0,},
};
MODULE_DEVICE_TABLE(pci, xi006_pci_tbl);

static irqreturn_t xi_interrupt(int irq, void * dev_id)
{
	int irqreturn = -1;
	XI006AEDevice *dev = (XI006AEDevice *)dev_id;
	
	if(XI006AEDevice_OnInterrupt(dev))
		irqreturn = IRQ_HANDLED;
	
	return IRQ_RETVAL(irqreturn);
}

static int /*__devinit*/ xi_pci_probe(struct pci_dev *pci, const struct pci_device_id *id)
{
	unsigned char *pbmmio;
	int ret = -ENOMEM;
	
	XI006AEDevice *dev = XI006AEDevice_Create();
	if(NULL == dev) { 
		ZH_DEBUG(0,"Error: XI006AEDevice create\n");
		goto fail_dev_create;
	}
	
	ret = pci_enable_device(pci);
	if (ret) {
		ZH_DEBUG(0,"Error: pci_enable_device\n");
		goto fail_pci_enable_device;
	}
	
	pci_set_master(pci);
	
	if(!request_mem_region(
						pci_resource_start(pci, 0), 
						pci_resource_len(pci, 0), 
						XI006AEPRO_INFO
						)) {
		ret = -EBUSY;
		ZH_DEBUG(0,"Error: request_mem_region\n");
		goto fail_release_mem;
	}
	
	pbmmio = ioremap_nocache(
						pci_resource_start(pci, 0), 
						pci_resource_len(pci, 0)
						);
	if(NULL == pbmmio) {
		ret = -EIO;
		ZH_DEBUG(0,"Error : ioremap_nocache \n");
		goto fail_ioremap;
	}
	
	ret = XI006AEDevice_OnPowerUp(dev,pbmmio,pci);
	if(ret) {
		ZH_DEBUG(0,"Error: XI006AEDevice init\n");
		goto fail_xi006aepro_init;
	}
	
	ret = request_irq(
					pci->irq, 
					xi_interrupt, 
					IRQF_SHARED,
					XI006AEPRO_INFO, 
					dev
					);
	if(ret) {
		ZH_DEBUG(0,"Error: request_irq\n");
		goto fail_free_irq;
	}
	
	pci_set_drvdata(pci, dev);
	
	xi_v4l2_init(dev);
	xi_alsa_init(dev);

	return 0;

fail_free_irq:
	pci->irq = -1;
	XI006AEDevice_OnPowerDown(dev);
	
fail_xi006aepro_init:
	iounmap(pbmmio);

fail_ioremap:
	release_mem_region(
				pci_resource_start(pci, 0), 
				pci_resource_len(pci, 0)
				);
	
fail_release_mem:
	pci_disable_device(pci);
	
fail_pci_enable_device:
	XI006AEDevice_Destroy(dev);
	
fail_dev_create:
	return ret;
}

void /*__devexit*/ xi_pci_remove(struct pci_dev *pci)
{
	unsigned char *pbmmio;
	XI006AEDevice *dev = pci_get_drvdata(pci);
	
	xi_alsa_release(dev);
	xi_v4l2_release(dev);

	pbmmio = XI006AEDevice_OnPowerDown(dev);
	
	if(pci->irq > 0) {
		free_irq(pci->irq, dev);
		pci->irq = -1;
	}

	if(pbmmio)
		iounmap(pbmmio);
		
	release_mem_region(
				pci_resource_start(pci, 0), 
				pci_resource_len(pci, 0)
				);

	pci_disable_device(pci);

	XI006AEDevice_Destroy(dev);
	
	printk(KERN_INFO "Remove XI006AEDevice successful \n");
}

#ifdef CONFIG_PM
static int xi_pci_suspend(struct pci_dev *pci, pm_message_t state)
{
	XI006AEDevice *dev = pci_get_drvdata(pci);
	
	ZH_DEBUG(0, "Enter XI100XE Suspend ..... \n");
	
	xi_alsa_suspend(dev);
	xi_v4l2_suspend(dev);
	XI006AEDevice_OnPowerDown(dev);
	
	if(pci->irq > 0) {
		free_irq(pci->irq, dev);
		pci->irq = -1;
	}
	
	pci_disable_device(pci);
	pci_save_state(pci);
	
	pci_set_power_state(pci, pci_choose_state(pci, state));
	
	ZH_DEBUG(0, "Lever XI100XE Suspend \n");
	
	return 0;
}

static int xi_pci_resume(struct pci_dev *pci)
{
	int ret; 
	XI006AEDevice *dev = pci_get_drvdata(pci);
	
	ZH_DEBUG(0, "Enter XI100XE Resume ..... \n");

	pci_set_power_state(pci, PCI_D0);
	pci_restore_state(pci);
	
	ret = pci_enable_device(pci);
	if(ret) {
		ZH_DEBUG(0,"Error: for resume on\n");
		ret = -EIO;
		goto out;
	}
	
	if(XI006AEDevice_OnPowerUp(dev,NULL,pci)){
		ZH_DEBUG(0,"Error: for resume on\n");
		ret = -EIO;
		goto out;
	}
	
	if(request_irq(pci->irq, xi_interrupt, IRQF_SHARED,"XI100XE", dev)){
		ZH_DEBUG(0,"Error: for resume on\n");
		ret = -EIO;
		goto out;
	}
	
	pci_set_master(pci);
	xi_alsa_resume(dev);
	xi_v4l2_resume(dev);
	ZH_DEBUG(0, "Lever XI100XE Resumed \n");
	
	return 0;
out:
	xi_pci_remove(pci);
	return ret;
}
#endif 

static struct pci_driver xi_pci_driver = {
	.name       	= XI006AEPRO_INFO,
	.id_table   	= xi006_pci_tbl,
	.probe      	= xi_pci_probe,
	.remove     	= /*__devexit_p*/(xi_pci_remove),
#ifdef CONFIG_PM
	.suspend 		= xi_pci_suspend,
	.resume 		= xi_pci_resume,
#endif 
};

static int __init xi_card_init(void)
{
	return pci_register_driver(&xi_pci_driver);
}

static void __exit xi_card_exit(void)
{
	pci_unregister_driver(&xi_pci_driver);
}

module_init(xi_card_init);
module_exit(xi_card_exit);
