
#include <linux/version.h>
#include <linux/init.h>
#include <linux/err.h>
#include <linux/jiffies.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/wait.h>
#include <linux/moduleparam.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/pcm.h>
#include <sound/rawmidi.h>
#include <sound/initval.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mutex.h>
#include <linux/kthread.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22)
#include <linux/freezer.h>
#endif

#include "Drivers/XI100XE_Device.h"
#include "alsa.h"

static unsigned int audio_devno;
static int index[SNDRV_CARDS] 	= SNDRV_DEFAULT_IDX;
static char *id[SNDRV_CARDS] 	= SNDRV_DEFAULT_STR;
static int enable[SNDRV_CARDS] 	= SNDRV_DEFAULT_ENABLE_PNP;

module_param_array(index, int, NULL, 0444);
MODULE_PARM_DESC(index, "Index value for audio soundcard.");
module_param_array(id, charp, NULL, 0444);
MODULE_PARM_DESC(id, "ID string for audio soundcard.");
module_param_array(enable, int, NULL, 0444);
MODULE_PARM_DESC(enable, "Enable this audio soundcard.");

/**************************************
			control
 **************************************/ 
static int volume_info(struct snd_kcontrol *kcontrol,struct snd_ctl_elem_info *info)
{
	info->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	info->count = 2;
	info->value.integer.min = 0;
	info->value.integer.max = 0x1F;
	return 0;
}

static int volume_line_get(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *value)
{
	XIAdrv *drv = snd_kcontrol_chip(kcontrol);
	value->value.integer.value[0] = drv->volume[0];
	value->value.integer.value[1] = drv->volume[1];
	return 0;
}

static int volume_line_put(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *value)
{
	XIAdrv *drv = snd_kcontrol_chip(kcontrol);
	
	uint8_t l = value->value.integer.value[0];
	uint8_t r = value->value.integer.value[1];
	if(l == drv->volume[0] && r == drv->volume[1])
		return 0;
		
	XI100XEDevice_SetALVolume(drv->dev, l);
	XI100XEDevice_SetARVolume(drv->dev, r);
	
	drv->volume[0] = l;
	drv->volume[1] = r;
	
	return 1;
}

static int volume_hdmi_get(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *value)
{
	XIAdrv *drv = snd_kcontrol_chip(kcontrol);
	value->value.integer.value[0] = drv->vhdmi[0];
	value->value.integer.value[1] = drv->vhdmi[1];
	return 0;
}

static int volume_hdmi_put(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *value)
{
	XIAdrv *drv = snd_kcontrol_chip(kcontrol);
	
	uint8_t l = value->value.integer.value[0];
	uint8_t r = value->value.integer.value[1];
	if(l == drv->vhdmi[0] && r == drv->vhdmi[1])
		return 0;
		
	XI100XEDevice_SetHDMILVolume(drv->dev, l);
	XI100XEDevice_SetHDMIRVolume(drv->dev, r);
	
	drv->vhdmi[0] = l;
	drv->vhdmi[1] = r;
	
	return 1;
}

static struct snd_kcontrol_new volume_kcontrol  = {
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE | SNDRV_CTL_ELEM_ACCESS_TLV_READ,
	.name = "Line Capture Volume",
	.info = volume_info,
	.get  = volume_line_get,
	.put  = volume_line_put,
};

static struct snd_kcontrol_new vhdmi_kcontrol  = {
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE | SNDRV_CTL_ELEM_ACCESS_TLV_READ,
	.name = "HDMI Capture Volume",
	.info = volume_info,
	.get  = volume_hdmi_get,
	.put  = volume_hdmi_put,
};

/**************************************
			thread
 **************************************/
 /*
  * lock by drv->lock
  */
static int audio_buffer_update(XIAdrv *drv, unsigned long frames)
 {
	int elapsed = 0;
	
	if (frames == 0)
		return elapsed;
		
	drv->ring_wpos_byframes += frames;
	drv->ring_wpos_byframes %= drv->ring_size_byframes;
	
	elapsed = (drv->period_used_byframes + frames) / drv->period_size_byframes;
	
	drv->period_used_byframes += frames;
	drv->period_used_byframes %= drv->period_size_byframes;
	
	return elapsed;
 }
 
static void alsa_harvest(XIAdrv *drv)
{
	int elapsed;
	unsigned long flags;
	unsigned long frames;
	struct snd_pcm_substream *substream = drv->substream;
	
	frames = XI100XEDevice_ACapture(drv->dev) / (2 * sizeof(short));

	spin_lock_irqsave(&drv->ring_lock, flags);
	elapsed = audio_buffer_update(drv,frames);
	spin_unlock_irqrestore(&drv->ring_lock, flags);
	
	if (elapsed && substream)
		snd_pcm_period_elapsed(substream);
}

static void audio_sleep(XIAdrv *drv,int jcount)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&drv->wq, &wait); 
	
	if ((!kthread_should_stop()) && (jcount > 0)) {
		schedule_timeout_interruptible(jcount);
	}
	remove_wait_queue(&drv->wq, &wait);	
}

static int audio_kthread(void *data)
{
	XIAdrv *drv = (XIAdrv *)data;
	
	ZH_DEBUG(0, "Start !\n");
	
	set_freezable();
	while(true) {
		if (kthread_should_stop())
			break;

		if(atomic_read(&drv->trigger)) 
			alsa_harvest(drv);
			
		audio_sleep(drv,HZ/100);
		try_to_freeze();
	}
	ZH_DEBUG(0, "Stop !\n");
	
	return 0;
}  
/**************************************
			file ops
 **************************************/
static struct snd_pcm_hardware audio_pcm_hardware = {
	.info 				=	(SNDRV_PCM_INFO_MMAP |
							SNDRV_PCM_INFO_INTERLEAVED |
							SNDRV_PCM_INFO_RESUME |
							SNDRV_PCM_INFO_MMAP_VALID),
	.formats 			=	SNDRV_PCM_FMTBIT_S16_LE, 
	.rates 				=	SNDRV_PCM_RATE_48000,
	.rate_min 			=	48000,
	.rate_max 			=	48000,
	.channels_min 		=	2,
	.channels_max 		=	2,
	.buffer_bytes_max 	=	64*1024,
	.period_bytes_min 	=	16*1024,
	.period_bytes_max 	=	64*1024,
	.periods_min 		=	4,
	.periods_max 		=	4,
};

/*
 * this callback is atomic. 
 * it cannot call that functions which may sleep. 
 * The trigger callback should be as minimal as possible
 * and this is the reason audio_kthread_xxxx remove away
 * 
 */
static int audio_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
	XIAdrv *drv = snd_pcm_substream_chip(substream);
	
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		ZH_DEBUG(0, "START\n");
		XI100XEDevice_AStart(
					drv->dev,
					substream->runtime->dma_area,
					substream->runtime->dma_bytes
					);
		//drv->ring_wpos_byframes = 0;
		//drv->period_used_byframes = 0;
		
		drv->ring_wpos_byframes = XI100XEDevice_AGetOSRingBufWPos(drv->dev) / (2 * sizeof(short));
		drv->period_used_byframes = drv->ring_wpos_byframes % drv->period_size_byframes;

	case SNDRV_PCM_TRIGGER_RESUME:
		ZH_DEBUG(0, "RESUME \n");
		atomic_set(&drv->trigger,1);
		wake_up_interruptible(&drv->wq);
		return 0;
		
	case SNDRV_PCM_TRIGGER_STOP:
		ZH_DEBUG(0, "STOP\n");
		XI100XEDevice_AStop(drv->dev);

	case SNDRV_PCM_TRIGGER_SUSPEND:	
		ZH_DEBUG(0, "SUSPEND\n");
		atomic_set(&drv->trigger,0);
		return 0;
		
	default: 
		return -EINVAL;
	}
}

/*
 * this  will be called each time snd_pcm_prepare() is called,
 * that this callback is now non-atomic
 * can use schedule-related functions safely in this callback
 * Be careful that this callback will be called many times at each setup
 */
static int audio_pcm_prepare(struct snd_pcm_substream *substream)
{
	unsigned long flags;
	XIAdrv *drv = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	
	spin_lock_irqsave(&drv->ring_lock, flags);
	drv->ring_size_byframes = runtime->buffer_size;
	drv->ring_wpos_byframes = 0;
	drv->period_size_byframes = runtime->period_size;
	drv->period_used_byframes = 0;
	spin_unlock_irqrestore(&drv->ring_lock, flags);
	
	ZH_DEBUG(0, "drv->ring_size_byframes is %d \n", drv->ring_size_byframes);
	ZH_DEBUG(0, "drv->period_size_byframes is %d \n", drv->period_size_byframes);
	
	return 0;
}

/*
 * This callback is called when the PCM middle layer 
 * inquires the current hardware position on the buffer.
 * The position must be returned in frames, ranging from 0 to buffer_size - 1.
 * This is called usually from the buffer-update routine in the pcm middle layer,
 * which is invoked when snd_pcm_period_elapsed() is called by timer
 * This callback is also atomic.
 */
static snd_pcm_uframes_t audio_pcm_pointer(struct snd_pcm_substream *substream)
{
	snd_pcm_uframes_t pos;
	XIAdrv *drv = snd_pcm_substream_chip(substream);
	
	spin_lock(&drv->ring_lock);
	pos = drv->ring_wpos_byframes;
	spin_unlock(&drv->ring_lock);
	
	return pos;
}

/*
 * This is called when the hardware parameter (hw_params) is set up by the application
 * the buffer size, the period size, the format, etc. are defined for the pcm substream
 * hardware setups should be done in this callback, including the allocation of buffers
 * Parameters to be initialized are retrieved by params_xxx() macros.
 * Note that this and prepare callbacks may be called multiple times per initialization.
 * this callback is non-atomic
 */
static int audio_pcm_hw_params(struct snd_pcm_substream *substream,struct snd_pcm_hw_params *hw_params)
{
	return snd_pcm_lib_malloc_pages(substream,params_buffer_bytes(hw_params));
}

/*
 * This is called to release the resources allocated via hw_params. 
 * This function is always called before the close callback is called
 * the callback may be called multiple times
 * Keep track whether the resource was already released
 */
static int audio_pcm_hw_free(struct snd_pcm_substream *substream)
{
	XIAdrv *drv = snd_pcm_substream_chip(substream);
		
	atomic_set(&drv->trigger,0);
	return snd_pcm_lib_free_pages(substream);
}

/*
 * This is called when a pcm substream is opened.
 * here have to initialize the runtime->hw record.
 * If the hardware configuration needs more constraints, set here
 */
static int audio_pcm_open(struct snd_pcm_substream *substream)
{
	XIAdrv *drv = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	
	runtime->hw = audio_pcm_hardware;
	drv->substream = substream;

	return 0;
}

/*
 * this is called when a pcm substream is closed
 * Any private instance for a pcm substream allocated 
 * in the open callback will be released here.
 */
static int audio_pcm_close(struct snd_pcm_substream *substream)
{
	return 0;
}

struct snd_pcm_ops audio_pcm_ops_no_buf = {
	.open 		=	audio_pcm_open,
	.close 		=	audio_pcm_close,
	.ioctl 		=	snd_pcm_lib_ioctl,
	.hw_params 	=	audio_pcm_hw_params,
	.hw_free 	=	audio_pcm_hw_free,
	.prepare 	=	audio_pcm_prepare,
	.trigger 	=	audio_pcm_trigger,
	.pointer 	=	audio_pcm_pointer,
};

/******************************************************************************
							module ops
******************************************************************************/

static int xi_audio_create(struct snd_card *card, XIAdrv **raudio)
{
	XIAdrv *drv = (XIAdrv *) card->private_data;
	memset(drv,0,sizeof(XIAdrv));

	drv->card = card;
	drv->volume[0] = 0x1F;
	drv->volume[1] = 0x1F;
	drv->vhdmi[0] = 0x1F;
	drv->vhdmi[1] = 0x1F;
	
	*raudio = drv;
	return 0;
}

static int snd_card_audio_pcm(XIAdrv *drv, int device,	int substreams)
{
	int err;
	struct snd_pcm *pcm;
	struct snd_pcm_ops *ops;
	
	err = snd_pcm_new(drv->card, "XI100XE", device, 0, substreams, &pcm);
	if (err < 0)
		return err;

	drv->pcm = pcm;
	ops = &audio_pcm_ops_no_buf;
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, ops);

	pcm->private_data = drv;
	pcm->info_flags = 0;
	strcpy(pcm->name, "XI100XE PCM");
	
	snd_pcm_lib_preallocate_pages_for_all(
							pcm,
							SNDRV_DMA_TYPE_CONTINUOUS,
							snd_dma_continuous_data(GFP_KERNEL),
							XI100XEDevice_ADefaultOSRingBufSize(drv->dev),//64*1024,
							XI100XEDevice_ADefaultOSRingBufSize(drv->dev)//64*1024,
							);
	
	return 0;
}

static int alsa_instance_create(void *dev,int inst)
{
	int err;
	struct snd_card  *card;
	XIAdrv  *drv;

	if (audio_devno >= SNDRV_CARDS)
		return (-ENODEV);

	if (!enable[audio_devno]) {
		++audio_devno;
		return (-ENOENT);
	}

	/*
	 * The function allocates snd_card instance via kzalloc 
	 * with the given space for the driver to use freely. 
	 * The allocated struct is stored in the given card_ret pointer.
	 */
	// snd_card_create() is deprecated as of 3.16
	// see http://forum.blackmagicdesign.com/viewtopic.php?f=12&t=26978
	#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 16, 0)
	err = snd_card_create(index[audio_devno], id[audio_devno], THIS_MODULE, sizeof(XIAdrv), &card);
	#else
	err = snd_card_new(dev, index[audio_devno], id[audio_devno], THIS_MODULE, sizeof(XIAdrv), &card);
	#endif
	
	if (err < 0)
		return err;

	err = xi_audio_create(card, &drv);
	if (err < 0)
		goto error;
		
	drv->channal = inst;
	drv->dev = dev;

	err = snd_card_audio_pcm(drv, inst, 1);
	if (err < 0)
		goto error;
	
	err = snd_ctl_add(card, snd_ctl_new1(&volume_kcontrol, drv));
	if (err < 0)
		goto error;
		
	ZH_DEBUG(0, "Add HDMI control ...... \n");
	err = snd_ctl_add(card, snd_ctl_new1(&vhdmi_kcontrol, drv));
	if (err < 0)
		goto error;
	
	ZH_DEBUG(0, "Add HDMI control OK \n");
	
	strcpy(card->driver, "XI100XE");
	strcpy(card->shortname, "XI100XE");
	sprintf(card->longname, "XI100XE %i", audio_devno);

	snd_card_set_dev(card, NULL);

	err = snd_card_register(card);
	if (err < 0)
		goto error;

	spin_lock_init(&drv->ring_lock);
	init_waitqueue_head(&drv->wq);
	atomic_set(&drv->trigger,0);
	
	//start kthread
	if(!drv->pthread) {
		drv->pthread = kthread_run(audio_kthread,drv,"XI100XE_ALSA_THREAD");
		if (IS_ERR(drv->pthread)) {
			err = PTR_ERR(drv->pthread);
			goto error;
		}
	}
	
	drv->romid = XI100XEDevice_GetROMID(dev,&drv->rlg);
	
	list_add_tail(&drv->list, XI100XE_GetAlsaListHead(dev));
	audio_devno++;
		
	return 0;
error:
	snd_card_free(card);
	return err;
}

int xi_alsa_release(void *dev)
{
	XIAdrv  *drv;
	struct list_head *list;
	struct list_head *head;
	
	head = XI100XE_GetAlsaListHead(dev);

	while (!list_empty(head)) {
		list = head->next;
		list_del(list);
		drv = list_entry(list, XIAdrv, list);
		if(drv->pthread) {
			kthread_stop(drv->pthread);
			drv->pthread = NULL;
		}
	
		snd_card_free(drv->card);
	}
	
	ZH_DEBUG(0, "Remove ALSA successful \n");
	
	return 0;
}

int xi_alsa_init(void *dev)
{
	int i, ret = -ENOMEM;

	for (i = 0; i < XI100XE_AUDIO_DEVICE_COUNT; i++) {
		ret = alsa_instance_create(dev,i);
		if (ret) 
			break;
	}
	ZH_DEBUG(0, "Create %d Capture Device\n",i);
	
	return ret;
}

void xi_alsa_suspend(void *dev)
{
	XIAdrv  *drv;
	struct list_head *pos;
	struct list_head *head;	
	
	ZH_DEBUG(0, "Enter ...\n");
	head = XI100XE_GetAlsaListHead(dev);
	
	list_for_each(pos, head) {
		drv = list_entry(pos, XIAdrv, list);
		
		snd_power_change_state(drv->card, SNDRV_CTL_POWER_D3hot);
		snd_pcm_suspend_all(drv->pcm);

		ZH_DEBUG(0, "Do one\n");
	}
	ZH_DEBUG(0, "Leave !\n");
}

void xi_alsa_resume(void *dev)
{
	XIAdrv  *drv;
	struct list_head *pos;
	struct list_head *head;	
	
	ZH_DEBUG(0, "Enter ...\n");
	
	head = XI100XE_GetAlsaListHead(dev);
	list_for_each(pos, head) {
		drv = list_entry(pos, XIAdrv, list);
		
		snd_power_change_state(drv->card, SNDRV_CTL_POWER_D0);
		
		ZH_DEBUG(0, "Do one\n");
	}
	ZH_DEBUG(0, "Leave !\n");
}
