
#include "OSAbstract/PrivateType.h"
#include "v4l2.h"

static VFmt formats[] = {
	{
		.name     = "4:2:2, packed, UYVY",
		.fourcc   = V4L2_PIX_FMT_UYVY,
		.depth    = 16,
	},
		{
		.name     = "4:2:2, packed, YUYV",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.depth    = 16,
	},
	{
		.name     = "RGB24",
		.fourcc   = V4L2_PIX_FMT_BGR24,
		.depth    = 24,
	},
	{
		.name     = "RGB32",
		.fourcc   = V4L2_PIX_FMT_BGR32,
		.depth    = 32,
	},
};
/*************************************************************
							misc ops
****************************************************************/
static VFmt *get_format(struct v4l2_format *f)
{
	unsigned int k;
		
	for (k = 0; k < ARRAY_SIZE(formats); k++) {
		if (formats[k].fourcc == f->fmt.pix.pixelformat)
			return &formats[k];
	}
	return NULL;
} 

/**************************************************************
 * 					kthread ops
 **************************************************************/ 
 static void v4l2_sleep(Vmanufactory *vm,unsigned long jcount)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&vm->wq, &wait);

	if ((!kthread_should_stop()) && (jcount > 0)) {
		schedule_timeout_interruptible(jcount);
	}
	remove_wait_queue(&vm->wq, &wait);
}

static void v4l2_framerate_adjust(Vmanufactory *vm)
{
	unsigned long jnext;
	
	if(!vm->framerate){
		vm->framerate = 25;
		return ;
	}
	
	if((!vm->prevrate) || (vm->prevrate != vm->framerate)){
		vm->prevrate = vm->framerate;
		vm->jbefore = jiffies;
		vm->jdelay = HZ/vm->framerate;
		return ;
	}
	
	jnext = vm->jbefore + vm->jdelay;
	if(time_is_after_jiffies(jnext)) {
		v4l2_sleep(vm,jnext - jiffies);
		vm->jbefore = jnext;
	} else {
		vm->jbefore = jiffies;
	}
}

static void v4l2_frame_syscby(Vmanufactory *vm, FrameSyscBy by)
{
	switch(by){
	case FRAME_SYSC_BY_IRQ:
		XI100XEDevice_WaitForNewImage(vm->dev, HZ/10);
		break;
	default:
		v4l2_framerate_adjust(vm);
	}
}

static void V4l2_ProcessOneFrame(Vmanufactory *vm)
{
	unsigned long flags = 0;
	VBuf *vbuf;
	
	spin_lock_irqsave(&vm->slock, flags);
	if (list_empty(&vm->actives)) {
		spin_unlock_irqrestore(&vm->slock, flags);
		return;
	}

	vbuf = list_entry(vm->actives.next, VBuf, buflist);
	list_del(&vbuf->buflist);
	spin_unlock_irqrestore(&vm->slock, flags);
	
	if(XI100XEDevice_ProcessOneFrame(vm->dev,vb2_plane_vaddr(&vbuf->vb,0))){
		do_gettimeofday(&vbuf->vb.v4l2_buf.timestamp);
		vbuf->vb.v4l2_buf.field = vm->field;
		vbuf->vb.v4l2_buf.sequence = vm->field_count++ ;
		vb2_buffer_done(&vbuf->vb,VB2_BUF_STATE_DONE);
	} else {
		do_gettimeofday(&vbuf->vb.v4l2_buf.timestamp);
		vb2_buffer_done(&vbuf->vb,VB2_BUF_STATE_ERROR);
	}
}

/**********************************************************
							extern ops
***********************************************************/
static int Vmanufactory_Start(Vmanufactory *vm,int(*threadfn)(void *data))
{
	XIVdrv *vdrv = container_of(vm, XIVdrv, vm);
	
	if(atomic_inc_return(&vdrv->used) > 1) {
		atomic_dec_return(&vdrv->used);
		return -EBUSY;
	}
	
	vm->field = V4L2_FIELD_ANY;
	vm->prevrate = 0;
	vm->field_count = 0;
	
	vm->pthread = kthread_run(threadfn, vm, "XI100XE_V4L2");
	if (IS_ERR(vm->pthread)) {
		return PTR_ERR(vm->pthread);
	}
	
	wake_up_interruptible(&vm->wq);

	return 0;
}

void Vmanufactory_Stop(Vmanufactory *vm)
{
	unsigned long flags = 0;
	XIVdrv *vdrv = container_of(vm, XIVdrv, vm);
	
	if (vm->pthread) {
		kthread_stop(vm->pthread);
		vm->pthread = NULL;
	}
	
	spin_lock_irqsave(&vm->slock, flags);
	while(!list_empty(&vm->actives)) {
		VBuf *vbuf = list_entry(vm->actives.next,
				VBuf, buflist);
		list_del_init(&vbuf->buflist);
		vb2_buffer_done(&vbuf->vb,VB2_BUF_STATE_ERROR);
	}
	spin_unlock_irqrestore(&vm->slock, flags);
	
	atomic_dec_return(&vdrv->used);
}

int v4l2_kthread_machine(void *data)
{
	Vmanufactory *vm = data;
	
	ZH_DEBUG(1,"Thread started\n");

	set_freezable();
	while(true) {
		if (kthread_should_stop())
			break;
		
		V4l2_ProcessOneFrame(vm);
		v4l2_frame_syscby(vm,FRAME_SYSC_BY_IRQ);

		try_to_freeze();
	}
	ZH_DEBUG(1,"Thread stoped\n");
	
	return 0;
}
/*********************************************************
							buffer ops
**********************************************************/
static int queue_setup(
					struct vb2_queue *vq, 
					const struct v4l2_format *fmt,
					unsigned int *nbuffers, 
					unsigned int *nplanes,
					unsigned int sizes[], 
					void *alloc_ctxs[]
					)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	unsigned long size = 	vm->show.cxWidth * \
							vm->show.cyHeight * \
							vm->show.depth >> 3;
	
	if (0 == *nbuffers)
		*nbuffers = 32;

	while (size * *nbuffers > VIDEO_BUF_MAX_M * 1024 * 1024)
		(*nbuffers)--;
		
	*nplanes = 1;
	sizes[0] = size;
	return 0;
}


static int buffer_init(struct vb2_buffer *vb)
{
	return 0;
}

static int buffer_prepare(struct vb2_buffer *vb)
{
	unsigned long size;
	Vmanufactory *vm = vb2_get_drv_priv(vb->vb2_queue);

	if (vm->show.cxWidth  < 48 || vm->show.cxWidth  > MAX_WIDTH ||
	    vm->show.cyHeight < 32 || vm->show.cyHeight > MAX_HEIGHT)
		return -EINVAL;

	size = vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;
	if (vb2_plane_size(vb, 0) < size) {
		return -EINVAL;
	}

	vb2_set_plane_payload(vb, 0, size);
	
	return 0;
}

static int buffer_finish(struct vb2_buffer *vb)
{
	return 0;
}

static void buffer_cleanup(struct vb2_buffer *vb)
{
	return ;
}

static void buffer_queue(struct vb2_buffer *vb)
{
	Vmanufactory *vm = vb2_get_drv_priv(vb->vb2_queue);
	VBuf *vbuf = container_of(vb, VBuf, vb);
	unsigned long flags = 0;

	spin_lock_irqsave(&vm->slock, flags);
	list_add_tail(&vbuf->buflist, &vm->actives);
	spin_unlock_irqrestore(&vm->slock, flags);
}

static int start_streaming(struct vb2_queue *vq, unsigned int count)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	return Vmanufactory_Start(vm,v4l2_kthread_machine);
}

static int stop_streaming(struct vb2_queue *vq)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	Vmanufactory_Stop(vm);
	return 0;
}

static void xi_lock(struct vb2_queue *vq)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	XIVdrv *drv = container_of(vm, XIVdrv, vm);
	mutex_lock(&drv->mutex);
}

static void xi_unlock(struct vb2_queue *vq)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	XIVdrv *drv = container_of(vm, XIVdrv, vm);
	mutex_unlock(&drv->mutex);
}

static struct vb2_ops xi_video_qops = {
	.queue_setup		= queue_setup,
	.buf_init			= buffer_init,
	.buf_prepare		= buffer_prepare,
	.buf_finish			= buffer_finish,
	.buf_cleanup		= buffer_cleanup,
	.buf_queue			= buffer_queue,
	.start_streaming	= start_streaming,
	.stop_streaming		= stop_streaming,
	.wait_prepare		= xi_unlock,
	.wait_finish		= xi_lock,
};

/******************************************************************************
							ioctl ops
******************************************************************************/
static int vidioc_querycap(struct file *file, void  *priv, struct v4l2_capability *cap)
{
	strcpy(cap->driver, "XI100XE");
	strcpy(cap->card, "XI100XE");
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE|
						V4L2_CAP_STREAMING    |
						V4L2_CAP_READWRITE;

	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv, struct v4l2_fmtdesc *f)
{
	VFmt *fmt;
	
	if (f->index >= ARRAY_SIZE(formats))
		return -EINVAL;

	fmt = &formats[f->index];

	strlcpy(f->description, fmt->name, sizeof(f->description));
	f->pixelformat = fmt->fourcc;

	return 0;
}

static int vidioc_g_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	f->fmt.pix.width        = vm->show.cxWidth;
	f->fmt.pix.height       = vm->show.cyHeight;
	f->fmt.pix.field        = vm->field;
	f->fmt.pix.pixelformat  = vm->show.fourcc;
	f->fmt.pix.bytesperline = (vm->show.cxWidth * 
								vm->show.cyHeight) >> 3;
	f->fmt.pix.sizeimage    = (vm->show.cxWidth * 
								vm->show.cyHeight * 
								vm->show.depth) >> 3;

	return 0;
}

static int vidioc_try_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	const VFmt *fmt;
	enum v4l2_field field;
	
	fmt = get_format(f);
	if (!fmt) {
		ZH_DEBUG(0,"Fourcc format (0x%08x) unknown.\n",f->fmt.pix.pixelformat);
		f->fmt.pix.pixelformat = formats[0].fourcc;
		fmt = get_format(f);
	}

	field = f->fmt.pix.field;
	if (field == V4L2_FIELD_ANY) {
		field = V4L2_FIELD_NONE;
	} else if (V4L2_FIELD_NONE != field) {
		return -EINVAL;
	}
	
	f->fmt.pix.field = field;
	v4l_bound_align_image(
					&f->fmt.pix.width, 
					48, 
					MAX_WIDTH, 
					2,
					&f->fmt.pix.height, 
					32, 
					MAX_HEIGHT, 
					0, 
					0
					);

	f->fmt.pix.bytesperline = (f->fmt.pix.width * fmt->depth) >> 3;
	f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;
		
	return 0;
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	int ret;
	VFmt *fmt = get_format(f);
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	struct vb2_queue *q = &vm->vb_vidq;	

	if (vb2_is_streaming(q)) {
		return -EBUSY;
	}

	ret = vidioc_try_fmt_vid_cap(file, priv, f);
	if (ret < 0) {
		ZH_DEBUG(0,"Invalid fmt!\n");
		return ret;
	}
	
	if(NULL == fmt)
		fmt = &formats[0];
			
	vm->show.cxWidth 	= f->fmt.pix.width;
	vm->show.cyHeight 	= f->fmt.pix.height;
	vm->show.csOutput 	= V4L2FMT2CSOUT(fmt->fourcc);
	vm->show.fourcc 	= fmt->fourcc;
	vm->show.depth 		= fmt->depth;
	vm->show.bBottomUp  = true;
	vm->field 			= f->fmt.pix.field;

	XI100XEDevice_SetVideoShowInfo(
				vm->dev,
				true, 
				V4L2FMT2CSOUT(vm->show.fourcc),
				vm->show.cxWidth,
				vm->show.cyHeight
				);

	return 0;
}

static int vidioc_reqbufs(struct file *file, void *priv, struct v4l2_requestbuffers *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	return vb2_reqbufs(&vm->vb_vidq, p);
}

static int vidioc_querybuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	return vb2_querybuf(&vm->vb_vidq, p);
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return vb2_qbuf(&vm->vb_vidq, p);
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
		
	return vb2_dqbuf(&vm->vb_vidq, p, file->f_flags & O_NONBLOCK);
}

static int vidioc_streamon(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return vb2_streamon(&vm->vb_vidq, i);
}

static int vidioc_streamoff(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return vb2_streamoff(&vm->vb_vidq, i);
}

static int vidioc_enum_input(struct file *file, void *priv, struct v4l2_input *inp)
{
	XIVdrv *vdrv = video_drvdata(file);
	if (inp->index != 0)
		return -EINVAL;

	inp->type = V4L2_INPUT_TYPE_CAMERA;
	strcpy(inp->name, "XI100XE");
	inp->std = vdrv->vfd->tvnorms;

	return 0;
}

static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
	*i = 0;
	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
	if (i > 0)
		return -EINVAL;

	return (0);
}

/******************************************************************************
							ioctl ops
******************************************************************************/
static int xi_g_volatile_ctrl(struct v4l2_ctrl *ctrl)
{
	return 0;
}

static int xi_s_ctrl(struct v4l2_ctrl *ctrl)
{
	return 0;
}

static ssize_t xi_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
	XIVdrv *drv = video_drvdata(file);

	return vb2_read(&drv->vm.vb_vidq, data, count, ppos, file->f_flags & O_NONBLOCK);
}

static unsigned int xi_poll(struct file *file, struct poll_table_struct *wait)
{
	XIVdrv *drv = video_drvdata(file);
	struct vb2_queue *q = &drv->vm.vb_vidq;
	
	return vb2_poll(q, file, wait);
}

static int xi_close(struct file *file)
{
	XIVdrv *drv = video_drvdata(file);

	if (v4l2_fh_is_singular_file(file))
		vb2_queue_release(&drv->vm.vb_vidq);
		
	return v4l2_fh_release(file);
}

static int xi_mmap(struct file *file, struct vm_area_struct *vma)
{
	XIVdrv *drv = video_drvdata(file);
	
	return vb2_mmap(&drv->vm.vb_vidq, vma);
}

static const struct v4l2_file_operations xi_fops = {
	.owner			= THIS_MODULE,
	.open           = v4l2_fh_open,
	.release        = xi_close,
	.read           = xi_read,
	.poll			= xi_poll,
	.unlocked_ioctl = video_ioctl2,
	.mmap           = xi_mmap,
};

static const struct v4l2_ioctl_ops xi_ioctl_ops = {
	.vidioc_querycap      		= vidioc_querycap,
	.vidioc_enum_fmt_vid_cap  	= vidioc_enum_fmt_vid_cap,
	.vidioc_g_fmt_vid_cap     	= vidioc_g_fmt_vid_cap,
	.vidioc_try_fmt_vid_cap   	= vidioc_try_fmt_vid_cap,
	.vidioc_s_fmt_vid_cap     	= vidioc_s_fmt_vid_cap,
	.vidioc_reqbufs       		= vidioc_reqbufs,
	.vidioc_querybuf      		= vidioc_querybuf,
	.vidioc_qbuf          		= vidioc_qbuf,
	.vidioc_dqbuf         		= vidioc_dqbuf,
	.vidioc_enum_input    		= vidioc_enum_input,
	.vidioc_g_input       		= vidioc_g_input,
	.vidioc_s_input       		= vidioc_s_input,
	.vidioc_streamon      		= vidioc_streamon,
	.vidioc_streamoff     		= vidioc_streamoff,
};

#define V4L2_STD_P_MASK 		(V4L2_STD_PAL_D | V4L2_STD_PAL_M)
#define V4L2_STD_M_MASK 		(V4L2_STD_NTSC_M | \
								V4L2_STD_NTSC_443 | \
								V4L2_STD_PAL_D | \
								V4L2_STD_PAL_M | \
								V4L2_STD_PAL_Nc | \
								V4L2_STD_PAL_60 | \
								V4L2_STD_SECAM_B | \
								V4L2_STD_SECAM_D)

static struct video_device xi_template = {
	.name			= "XI100XE",
	.fops        	= &xi_fops,
	.ioctl_ops 		= &xi_ioctl_ops,
	.release		= video_device_release,

	.tvnorms  		= V4L2_STD_M_MASK,
	//.current_norm 	= V4L2_STD_NTSC_M,
};
/******************************************************************************
							module ops
******************************************************************************/
static const struct v4l2_ctrl_ops xi_ctrl_ops = {
	.g_volatile_ctrl = xi_g_volatile_ctrl,
	.s_ctrl 		 = xi_s_ctrl,
};

/*
static const struct v4l2_ctrl_config xi_ctrl_squaremode = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SQUAREMODE,
	.name = "Use Square Mode",
	.type = V4L2_CTRL_TYPE_BOOLEAN,
	.min = 0,
	.max = 1,
	.step = 1,
	.def = 0,
};

static const struct v4l2_ctrl_config xi_ctrl_agcmode = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_AGCMODE,
	.name = "Use AGC Mode",
	.type = V4L2_CTRL_TYPE_INTEGER,
	.min = 0,
	.max = 255,
	.step = 1,
	.def = 0,//0x0,0x01,0x10,0x11
};

static const char * const xi_ctrl_deinterlace_menu[] = {
	"DEINTERLACE_BLEND",
	"DEINTERLACE_NONE",
	NULL,
};
    
static const struct v4l2_ctrl_config xi_ctrl_deinterlace = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_DEINTERLACE,
	.name = "Swith Deinterlace",
	.type = V4L2_CTRL_TYPE_MENU,
	.min = 0,
	.max = 2,
	.def = 1,
	//.menu_skip_mask = 0x04,
	.qmenu = xi_ctrl_deinterlace_menu,
};

static const char * const xi_ctrl_scale_menu[] = {
	"SCALEP_TYPE_SQEEZE ",
	"SCALEP_TYPE_LETTER_BOX",
	"SCALEP_TYPE_EDGE_CROP",
	NULL,
};

static const struct v4l2_ctrl_config xi_ctrl_scale = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SCALE,
	.name = "Swith Scale Mode",
	.type = V4L2_CTRL_TYPE_MENU,
	.min = 0,
	.max = 3,
	.def = 0,
	//.menu_skip_mask = 0x04,
	.qmenu = xi_ctrl_scale_menu,
};

static const struct v4l2_ctrl_config xi_ctrl_vflip = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_VFLIP,
	.name = "Swith On/Off Vflip",
	.type = V4L2_CTRL_TYPE_BOOLEAN,
	.min = 0,
	.max = 1,
	.step = 1,
	.def = 0,
};

static const struct v4l2_ctrl_config xi_ctrl_hflip = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_HFLIP,
	.name = "Swith On/Off Hflip",
	.type = V4L2_CTRL_TYPE_BOOLEAN,
	.min = 0,
	.max = 1,
	.step = 1,
	.def = 0,
};

static const struct v4l2_ctrl_config xi_ctrl_signalvalid = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SIGNALVALID,
	.name = "Test Signal Valid/Invalid",
	.type = V4L2_CTRL_TYPE_BOOLEAN,
	.min = 0,
	.max = 1,
	.step = 1,
	.def = 0,
};

static const struct v4l2_ctrl_config xi_ctrl_signalchanged = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SIGNALCHANGED,
	.name = "Test Signal Changed",
	.type = V4L2_CTRL_TYPE_BOOLEAN,
	.min = 0,
	.max = 1,
	.step = 1,
	.def = 0,
};

static const struct v4l2_ctrl_config xi_ctrl_srcw = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SIGNALWIDTH,
	.name = "Image Source Width",
	.type = V4L2_CTRL_TYPE_INTEGER,
	.min = 64,//640,
	.max = 1920,//768,
	.step = 16,
	.def = 768,
};

static const struct v4l2_ctrl_config xi_ctrl_srch = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SIGNALHEIGHT,
	.name = "Image Source Height",
	.type = V4L2_CTRL_TYPE_INTEGER,
	.min = 32,//480,
	.max = 1080,//576,
	.step = 16,
	.def = 576,
};

static const struct v4l2_ctrl_config xi_ctrl_srcfps = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_SIGNALFPS,
	.name = "Image Source FPS",
	.type = V4L2_CTRL_TYPE_INTEGER,
	.min = 1,
	.max = 30,
	.step = 1,
	.def = 25,
};
*/

static int __init v4l2_instance_create(void *dev, int inst)
{
	int ret;
	XIVdrv *vdrv;
	struct video_device *vfd;
	struct vb2_queue *q;

	vdrv = kzalloc(sizeof(*vdrv), GFP_KERNEL);
	if (!vdrv)
		return -ENOMEM;

	snprintf(vdrv->v4l2_dev.name, sizeof(vdrv->v4l2_dev.name),"%s-%03d", "XI100XE", inst);
	ret = v4l2_device_register(NULL, &vdrv->v4l2_dev);
	if (ret)
		goto v4l2_reg_fail;

	vdrv->vm.show.cxWidth  = 768;
	vdrv->vm.show.cyHeight = 576;
	vdrv->vm.show.csOutput = V4L2FMT2CSOUT(V4L2_PIX_FMT_UYVY);
	vdrv->vm.show.fourcc = V4L2_PIX_FMT_UYVY;
	vdrv->vm.show.depth = 16;
	vdrv->vm.show.bBottomUp = true;
	
	INIT_LIST_HEAD(&vdrv->vm.actives);
	init_waitqueue_head(&vdrv->vm.wq);
	vdrv->vm.channal = inst;
	vdrv->vm.dev = dev;
	vdrv->vm.jbefore = 0;
	vdrv->vm.jdelay = 0;
	vdrv->vm.prevrate = 0;
	vdrv->vm.framerate = 25;
	spin_lock_init(&vdrv->vm.slock);
	vdrv->vm.field = V4L2_FIELD_NONE;
	vdrv->vm.field_count = 0;
	
	q = &vdrv->vm.vb_vidq;
	memset(q, 0, sizeof(vdrv->vm.vb_vidq));
	q->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	q->io_modes = VB2_MMAP | VB2_USERPTR | VB2_READ;
	q->drv_priv = &vdrv->vm;
	q->buf_struct_size = sizeof(VBuf);
	q->ops = &xi_video_qops;
	q->mem_ops = &vb2_vmalloc_memops;
	vb2_queue_init(q);

	mutex_init(&vdrv->mutex);
	atomic_set(&vdrv->used,0);
	vdrv->channal = inst;
	vdrv->dev = dev;
	
	vfd = video_device_alloc();
	if (!vfd) {
		ret = -ENOMEM;
		goto vfd_alloc_fail;
	}
	*vfd = xi_template;
	vfd->debug = 0;
	vfd->v4l2_dev = &vdrv->v4l2_dev;
	set_bit(V4L2_FL_USES_V4L2_FH, &vfd->flags);
	vfd->lock = &vdrv->mutex;
	ret = video_register_device(vfd, VFL_TYPE_GRABBER, -1);
	if (ret < 0)
		goto vfd_reg_fail;

	video_set_drvdata(vfd, vdrv);
	vdrv->vfd = vfd;

	list_add_tail(&vdrv->vlist, XI100XE_GetV4l2ListHead(dev));
	return 0;

vfd_reg_fail:
	video_device_release(vfd);
vfd_alloc_fail:
	v4l2_device_unregister(&vdrv->v4l2_dev);
v4l2_reg_fail:
	kfree(vdrv);
	return ret;
}

int xi_v4l2_release(void *dev)
{
	struct list_head *head = XI100XE_GetV4l2ListHead(dev);
	
	while (!list_empty(head)) {
		XIVdrv *vdrv = list_entry(head->next, XIVdrv, vlist);
		list_del(&vdrv->vlist);

		v4l2_ctrl_handler_free(&vdrv->ctrl_handler);
		video_unregister_device(vdrv->vfd);
		v4l2_device_unregister(&vdrv->v4l2_dev);
		kfree(vdrv);
	}

	return 0;
}

int xi_v4l2_init(void *dev)
{
	int ret = -ENOMEM, i;

	for (i = 0; i < XI100XE_VIDEO_DEVICE_COUNT; i++) {
		ret = v4l2_instance_create(dev,i);
		if (ret) 
			break;
	}
	ZH_DEBUG(1, "video counts = %d\n",i);
	return ret;
}
