
#include "v4l2_centos5.h"

static VFmt formats[] = {
	{
		.name     = "4:2:2, packed, UYVY",
		.fourcc   = V4L2_PIX_FMT_UYVY,
		.depth    = 16,
	},
		{
		.name     = "4:2:2, packed, YUYV",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.depth    = 16,
	},
	{
		.name     = "RGB24",
		.fourcc   = V4L2_PIX_FMT_BGR24,
		.depth    = 24,
	},
	{
		.name     = "RGB32",
		.fourcc   = V4L2_PIX_FMT_BGR32,
		.depth    = 32,
	},
};

#define XI_CID_CUSTOM_BASE			(V4L2_CID_USER_BASE | 0xe000)
#define XI_CID_CUSTOM_VFLIP  		(XI_CID_CUSTOM_BASE + 1)
#define XI_CID_CUSTOM_HFLIP			(XI_CID_CUSTOM_BASE + 2)
#define XI_CID_CUSTOM_AUTO_TRIGGER	(XI_CID_CUSTOM_BASE + 3)
#define XI_CID_CUSTOM_DEINTERLACE	(XI_CID_CUSTOM_BASE + 4)
#define XI_CID_CUSTOM_SCALE			(XI_CID_CUSTOM_BASE + 5)
#define XI_CID_CUSTOM_PHASE			(XI_CID_CUSTOM_BASE + 6)
#define XI_CID_CUSTOM_TIMING		(XI_CID_CUSTOM_BASE + 7)

const static struct v4l2_queryctrl xi_qctrl[] = {
	{
		.id            = V4L2_CID_BRIGHTNESS,
		.type          = V4L2_CTRL_TYPE_INTEGER,
		.name          = "Brightness",
		.minimum       = -128,
		.maximum       = 127,
		.step          = 1,
		.default_value = 10,
		.flags         = 0,
	}, {
		.id            = V4L2_CID_CONTRAST,
		.type          = V4L2_CTRL_TYPE_INTEGER,
		.name          = "Contrast",
		.minimum       = 0,
		.maximum       = 255,
		.step          = 1,
		.default_value = 128,
		.flags         = 0,
	}, {
		.id            = V4L2_CID_SATURATION,
		.type          = V4L2_CTRL_TYPE_INTEGER,
		.name          = "Saturation",
		.minimum       = 0,
		.maximum       = 255,
		.step          = 0x1,
		.default_value = 128,
		.flags         = 0,
	}, {
		.id            = V4L2_CID_HUE,
		.type          = V4L2_CTRL_TYPE_INTEGER,
		.name          = "Hue",
		.minimum       = -128,
		.maximum       = 127,
		.step          = 0x1,
		.default_value = 0,
		.flags         = 0,
	}, {
		.id 			= XI_CID_CUSTOM_HFLIP,
		.type 			= V4L2_CTRL_TYPE_BOOLEAN,
        .name 			= "Image HFLIP",
		.minimum 		= 0,
		.maximum 		= 1,
		.step 			= 1,
		.default_value 	= 0,
	}, {
		.id 			= XI_CID_CUSTOM_VFLIP,
		.type 			= V4L2_CTRL_TYPE_BOOLEAN,
        .name 			= "Image VFLIP",
		.minimum 		= 0,
		.maximum 		= 1,
		.step 			= 1,
		.default_value 	= 0,
	}, {
		.id            	= XI_CID_CUSTOM_AUTO_TRIGGER,
		.type          	= V4L2_CTRL_TYPE_BOOLEAN,
		.name          	= "AUTO TRIGGER",
		.minimum 		= 0,
		.maximum 		= 1,
		.step 			= 1,
		.default_value 	= 0,	
	}, {
		.id       		= XI_CID_CUSTOM_DEINTERLACE,
		.type          	= V4L2_CTRL_TYPE_MENU,
		.name          	= "DEINTERLACE",
		.minimum       	= 0,
		.maximum       	= 1,
		.step          	= 0x1,
		.default_value 	= 0,
	}, {
		.id            = XI_CID_CUSTOM_SCALE,
		.type          = V4L2_CTRL_TYPE_MENU,
		.name          = "SCALE",
		.minimum       = 0,
		.maximum       = 2,
		.step          = 0x1,
		.default_value = 0,
	}, {
		.id            = XI_CID_CUSTOM_PHASE,
		.type          = V4L2_CTRL_TYPE_INTEGER,
		.name          = "PHASE",
		.minimum       = 0,
		.maximum       = 0x1F,
		.step          = 0x1,
		.default_value = 0,
	}, {
		.id            = XI_CID_CUSTOM_TIMING,
		.type          = V4L2_CTRL_TYPE_INTEGER,
		.name          = "TIMING",
		.minimum       = 0,
		.maximum       = 255,
		.step          = 0x1,
		.default_value = 0,
	},
};

/******************************************************************************
							misc ops
******************************************************************************/
static VFmt *get_format(struct v4l2_format *f)
{
	unsigned int k;
		
	for (k = 0; k < ARRAY_SIZE(formats); k++) {
		if (formats[k].fourcc == f->fmt.pix.pixelformat)
			return &formats[k];
	}
	return NULL;
} 

#define time_is_after_jiffies(time) time_before(jiffies, time)
/**************************************************************
 * 					kthread ops
 **************************************************************/ 
 static void v4l2_sleep(Vmanufactory *vm, unsigned long jcount)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&vm->wq, &wait);

	if ((!kthread_should_stop()) && (jcount > 0)) {
		schedule_timeout_interruptible(jcount);
	}

	remove_wait_queue(&vm->wq, &wait);
}

static void v4l2_framerate_adjust(Vmanufactory *vm)
{
	unsigned long jnext;
	
	if(!vm->framerate){
		vm->framerate = 25;
		return ;
	}
	
	if((!vm->prevrate) || (vm->prevrate != vm->framerate)){
		vm->prevrate = vm->framerate;
		vm->jbefore = jiffies;
		vm->jdelay = HZ/vm->framerate;
		return ;
	}
	
	jnext = vm->jbefore + vm->jdelay;
	if(time_is_after_jiffies(jnext)) {
		v4l2_sleep(vm,jnext - jiffies);
		vm->jbefore = jnext;
	} else {
		vm->jbefore = jiffies;
	}
}

static void v4l2_frame_syscby(Vmanufactory *vm, FrameSyscBy by)
{
	switch(by){
	case FRAME_SYSC_BY_IRQ:
		XI102XEHD_WaitForNewImage(vm->dev, vm->channal, HZ/10);
		break;
	default:
		v4l2_framerate_adjust(vm);
	}
}

static void framebuf_to_baddr(struct videobuf_buffer *vb, unsigned char  *data, int size)
{
	int i;
	struct page *pg;
	unsigned char * vaddr;

	for(i = 0; i < vb->dma.nr_pages; i++) {
		pg = pfn_to_page(vb->dma.sglist[i].dma_address >> PAGE_SHIFT);
		vaddr = kmap_atomic(pg,KM_BOUNCE_READ) + vb->dma.sglist[i].offset;
		memcpy(vaddr, data, vb->dma.sglist[i].length);
		kunmap_atomic(vaddr, KM_BOUNCE_READ);
		size -= vb->dma.sglist[i].length;
		data += vb->dma.sglist[i].length;
		if(size <= 0)
			return ;
	}
}

static void V4l2_ProcessOneFrame(Vmanufactory *vm)
{
	unsigned long flags = 0;
	struct videobuf_buffer *vb;

	spin_lock_irqsave(&vm->slock, flags);
	if (list_empty(&vm->actives)) {
		spin_unlock_irqrestore(&vm->slock, flags);
		return;
	}

	vb = list_entry(vm->actives.next, struct videobuf_buffer, queue);
	if (!waitqueue_active(&vb->done)){
		spin_unlock_irqrestore(&vm->slock, flags);
		return;
	}

	list_del(&vb->queue);
	spin_unlock_irqrestore(&vm->slock, flags);
	
	if((vm->show.cxWidth != vb->width) || (vm->show.cyHeight != vb->height) || (vm->frame_buf == NULL)) {
		vb->state = STATE_ERROR;
	} else {
		XI102XEHD_ProcessOneFrame(vm->dev,vm->channal, vm->frame_buf);
		framebuf_to_baddr(vb, vm->frame_buf, vm->frame_size);
		vb->state = STATE_DONE;
	}
	
	vb->field_count++;
	wake_up(&vb->done);

	return;
}
/******************************************************************************
							extern ops
******************************************************************************/
int Vmanufactory_Start(Vmanufactory *vm,int(*threadfn)(void *data))
{
	vm->pthread = kthread_run(threadfn, vm, "XI100XE_V4L2_THREAD");
	if (IS_ERR(vm->pthread)) {
		return PTR_ERR(vm->pthread);
	}
	
	wake_up_interruptible(&vm->wq);
	
	return 0;
}

void Vmanufactory_Stop(Vmanufactory *vm)
{
	unsigned long flags = 0;

	if (vm->pthread)
		kthread_stop(vm->pthread);

	vm->pthread = NULL;
	
	spin_lock_irqsave(&vm->slock, flags);
	while(!list_empty(&vm->actives)) {
		struct videobuf_buffer *vb;
		vb = list_entry(vm->actives.next, struct videobuf_buffer, queue);
		list_del(&vb->queue);
		vb->state = STATE_ERROR;
		wake_up(&vb->done);
	}
	spin_unlock_irqrestore(&vm->slock, flags);
}

int v4l2_kthread_machine(void *data)
{
	Vmanufactory *vm = data;
	
	while(true) {
		if (kthread_should_stop())
			break;
		
		if(vm->show.cxWidth)
			V4l2_ProcessOneFrame(vm);
			
		v4l2_frame_syscby(vm,FRAME_SYSC_BY_IRQ);
	}

	return 0;
}

/******************************************************************************
							buffer ops
******************************************************************************/
static int v4l2_buffer_setup(struct videobuf_queue *vq, unsigned int *count, unsigned int *size)
{
	Vmanufactory *vm = vq->priv_data;
	*size = vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;

	if (0 == *count)
		*count = 32;

	while (*size * *count > VIDEO_BUF_MAX_M * 1024 * 1024)
		(*count)--;
	
	return 0;
}

static void v4l2_buffer_release(struct videobuf_queue *vq, struct videobuf_buffer *vb)
{
	if (in_interrupt())
		BUG();

	videobuf_waiton(vb,0,0);
	videobuf_dma_unmap(vq, &vb->dma);
	videobuf_dma_free(&vb->dma);

	vb->state = STATE_NEEDS_INIT;
}

static int v4l2_buffer_prepare(struct videobuf_queue *vq, struct videobuf_buffer *vb, enum v4l2_field field)
{
	int rc;
	Vmanufactory *vm = vq->priv_data;

	if (vm->show.cxWidth  < 48 || vm->show.cxWidth  > 1920 || vm->show.cyHeight < 32 || vm->show.cyHeight > 1080)
		return -EINVAL;

	vb->size =  vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;
			
	if (0 != vb->baddr && vb->bsize < vb->size)
		return -EINVAL;
		
	vb->width  = vm->show.cxWidth;
	vb->height = vm->show.cyHeight;
	vb->bytesperline = vm->show.cxWidth * vm->show.depth >> 3;
	vb->field  = field;
	
	if (STATE_NEEDS_INIT == vb->state) {
		rc = videobuf_iolock(vq, vb, NULL);
		if (rc < 0)
			goto fail;
	}

	vb->state = STATE_PREPARED;
	return 0;

fail:
	v4l2_buffer_release(vq, vb);
	return rc;
}

static void v4l2_buffer_queue(struct videobuf_queue *vq, struct videobuf_buffer *vb)
{
	Vmanufactory *vm = vq->priv_data;
	
	vb->state = STATE_QUEUED;
	list_add_tail(&vb->queue, &vm->actives);
}

/******************************************************************************
							ioctl ops
******************************************************************************/
static int vidioc_querycap(struct file *file, void  *priv, struct v4l2_capability *cap)
{
	strcpy(cap->driver, "XI100XE");
	strcpy(cap->card, "XI100XE");
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE| V4L2_CAP_STREAMING|V4L2_CAP_READWRITE;

	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv,struct v4l2_fmtdesc *f)
{
	const VFmt *fmt;
	
	if (f->index >= ARRAY_SIZE(formats))
		return -EINVAL;

	fmt = &formats[f->index];

	strlcpy(f->description, fmt->name, sizeof(f->description));
	f->pixelformat = fmt->fourcc;

	return 0;
}

static int vidioc_g_fmt_vid_cap(struct file *file, void *priv,struct v4l2_format *f)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;

	f->fmt.pix.width        = vm->show.cxWidth;
	f->fmt.pix.height       = vm->show.cyHeight;
	f->fmt.pix.field        = vm->vb_vidq.field;
	f->fmt.pix.pixelformat  = vm->show.fourcc;
	f->fmt.pix.bytesperline = (vm->show.cxWidth * vm->show.depth) >> 3;
	f->fmt.pix.sizeimage 	= (vm->show.cxWidth * vm->show.cyHeight * vm->show.depth) >> 3;
	return (0);
}

static int vidioc_try_fmt_vid_cap(struct file *file, void *priv,struct v4l2_format *f)
{
	const VFmt *fmt;
	enum v4l2_field field;

	fmt = get_format(f);
	if (!fmt) {
		printk(KERN_WARNING "Format Mismatch!\n");
		f->fmt.pix.pixelformat = formats[0].fourcc;
		return -EINVAL;
	}

	field = f->fmt.pix.field;
	if (field == V4L2_FIELD_ANY) {
		field = V4L2_FIELD_NONE;
	} else if (V4L2_FIELD_NONE != field) {
		return -EINVAL;
	}

	f->fmt.pix.field = field;
	if (f->fmt.pix.height < 32)f->fmt.pix.height = 32;
	if (f->fmt.pix.height > MAX_HEIGHT)f->fmt.pix.height = MAX_HEIGHT;
	if (f->fmt.pix.width < 48)f->fmt.pix.width = 48;
	if (f->fmt.pix.width > MAX_WIDTH)f->fmt.pix.width = MAX_WIDTH;
	f->fmt.pix.width &= ~0x03;
	f->fmt.pix.bytesperline = (f->fmt.pix.width * fmt->depth) >> 3;
	f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;

	return 0;
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	unsigned int size;
	VFmt *fmt = get_format(f);
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;
	struct videobuf_queue *q = &vm->vb_vidq;

	int ret = vidioc_try_fmt_vid_cap(file, drv, f);
	if (ret < 0) {
		printk(KERN_WARNING "Can't set fmt!\n");
		return (ret);
	}
	
	if (videobuf_queue_is_busy(q)) {
		ret = -EBUSY;
		goto out;
	}

	if(NULL == fmt)
		fmt = &formats[0];
	
	mutex_lock(&q->vb_lock);
	vm->show.fourcc = fmt->fourcc;
	vm->show.depth = fmt->depth;
	vm->show.cxWidth = f->fmt.pix.width;
	vm->show.cyHeight = f->fmt.pix.height;
	vm->vb_vidq.field = f->fmt.pix.field;
	vm->type = f->type;
	
	size = vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;

	if(NULL == vm->frame_buf) {
		vm->frame_size = size;
		vm->frame_buf = vmalloc_32(size);
		if (vm->frame_buf == NULL) {
			ret = -ENOMEM;
			goto out;
		}
	}

	if(size != vm->frame_size) {
		vfree(vm->frame_buf);
		vm->frame_size = size;
		vm->frame_buf = vmalloc_32(size);
		if (vm->frame_buf == NULL) {
			ret = -ENOMEM;
			goto out;
		}
	}
	
	XI102XEHD_SetVideoShowInfo(
					vm->dev,
					vm->channal,
					true, 
					V4L2FMT2CSOUT(vm->show.fourcc),
					vm->show.cxWidth,
					vm->show.cyHeight
					);

	ret = 0;
out:
	return ret;
}

static int vidioc_reqbufs(struct file *file, void *priv, struct v4l2_requestbuffers *p)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;

	return (videobuf_reqbufs(&vm->vb_vidq, p));
}

static int vidioc_querybuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;

	return (videobuf_querybuf(&vm->vb_vidq, p));
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;

	return (videobuf_qbuf(&vm->vb_vidq, p));
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;
		
	return (videobuf_dqbuf(&vm->vb_vidq, p, file->f_flags & (O_NONBLOCK)));
}

static int vidioc_enum_input(struct file *file, void *priv,struct v4l2_input *inp)
{
	if (inp->index != 0)
		return -EINVAL;

	inp->type = V4L2_INPUT_TYPE_CAMERA;
	strcpy(inp->name, "XI100XE");
	inp->std = V4L2_STD_NTSC_M;

	return 0;
}

static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
	*i = 0;
	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
	if (i > 0)
		return -EINVAL;	

	return (0);
}

static int vidioc_queryctrl(struct file *file, void *priv, struct v4l2_queryctrl *qc)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(xi_qctrl); i++)
		if (qc->id && qc->id == xi_qctrl[i].id){
			memcpy(qc, &(xi_qctrl[i]),sizeof(*qc));
			return (0);
		}

	return -EINVAL;
}

static int vidioc_g_ctrl(struct file *file, void *priv, struct v4l2_control *ctrl)
{
	int i;
	XIVdrv *vdrv = priv;

	for (i = 0; i < ARRAY_SIZE(xi_qctrl); i++)
		if (ctrl->id == xi_qctrl[i].id) {
			ctrl->value = vdrv->configure[i];
			return (0);
		}

	return -EINVAL;
}

static int vidioc_s_ctrl(struct file *file, void *priv, struct v4l2_control *ctrl)
{
	return 0;
}

static int vidioc_streamon(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;

	if (vm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (i != vm->type)
		return -EINVAL;

	return videobuf_streamon(&vm->vb_vidq);
}

static int vidioc_streamoff(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = priv;
	Vmanufactory *vm = &drv->vm;

	if (vm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		ZH_DEBUG(0,"Stream off Error\n");
		return -EINVAL;
	}
	if (i != vm->type) {
		ZH_DEBUG(0,"Stream off type Error\n");
		return -EINVAL;
	}

	return videobuf_streamoff(&vm->vb_vidq);
}

static int xi_map_sg(void *dev, struct scatterlist *sg, int nents, int direction)
{
	int i;

	BUG_ON(direction == DMA_NONE);

	for (i = 0; i < nents; i++ ) {
		BUG_ON(!sg[i].page);

		sg[i].dma_address = page_to_phys(sg[i].page) + sg[i].offset;
	}

	return nents;
}

static int xi_unmap_sg(void *dev,struct scatterlist *sglist,int nr_pages, int direction)
{
	return 0;
}

static int xi_dma_sync_sg(void *dev,struct scatterlist *sglist, int nr_pages, int direction)
{
	return 0;
}

/******************************************************************************
							fops ops
******************************************************************************/
static struct videobuf_queue_ops xi_video_qops = {
	.buf_setup      = v4l2_buffer_setup,
	.buf_prepare    = v4l2_buffer_prepare,
	.buf_queue      = v4l2_buffer_queue,
	.buf_release    = v4l2_buffer_release,
	.vb_map_sg      = xi_map_sg,
	.vb_dma_sync_sg = xi_dma_sync_sg,
	.vb_unmap_sg    = xi_unmap_sg,
};

static int xi_open(struct inode *inode, struct file *file)
{
	struct video_device *dev = video_devdata(file);

	XIVdrv	*drv 	= dev->priv;
	
	mutex_lock(&drv->mutex);
	if(drv->used) {
		mutex_unlock(&drv->mutex);
		return -EBUSY;
	} else {
		drv->used++;
		mutex_unlock(&drv->mutex);
	}
	
	drv->vm.framerate = 25;
	drv->vm.show.cxWidth = 1080;
	drv->vm.show.cyHeight = 720;
	drv->vm.show.fourcc = V4L2_PIX_FMT_YUYV;
	file->private_data = drv;
	
	videobuf_queue_init(
				&drv->vm.vb_vidq, 
				&xi_video_qops,
				NULL, 
				&drv->vm.slock,
				drv->vm.type, 
				V4L2_FIELD_NONE,
				sizeof(struct videobuf_buffer), 
				&drv->vm
				);

	return Vmanufactory_Start(&drv->vm,v4l2_kthread_machine);
}

static int xi_close(struct inode *inode, struct file *file)
{
	XIVdrv	*v4l2 = file->private_data;
	Vmanufactory *vm = &v4l2->vm;

	Vmanufactory_Stop(vm);

	videobuf_mmap_free(&vm->vb_vidq);

	mutex_lock(&v4l2->mutex);
	v4l2->used = 0;
	mutex_unlock(&v4l2->mutex);

	if(vm->frame_buf) {
		vfree(vm->frame_buf);
		vm->frame_size = 0;
		vm->frame_buf = NULL;
	}
	
	return 0;
}

static ssize_t xi_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
	XIVdrv	*v4l2 = file->private_data;

	if (v4l2->vm.type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return videobuf_read_one(&v4l2->vm.vb_vidq, data, count, ppos, file->f_flags & O_NONBLOCK);
	}
	return 0;
}

static unsigned int xi_poll(struct file *file, struct poll_table_struct *wait)
{
	XIVdrv	*v4l2 = file->private_data;
	struct videobuf_queue *q = &v4l2->vm.vb_vidq;
	struct videobuf_buffer *vb;

	if (V4L2_BUF_TYPE_VIDEO_CAPTURE != v4l2->vm.type)
		return POLLERR;

	if (list_empty(&q->stream))
		return POLLERR;

	vb = list_entry(q->stream.next, struct videobuf_buffer, stream);

	poll_wait(file, &vb->done, wait);
	if (vb->state == STATE_DONE || vb->state == STATE_ERROR)
		return POLLIN|POLLRDNORM;
		
	return 0;
}

static int xi_mmap(struct file *file, struct vm_area_struct *vma)
{
	XIVdrv	*v4l2 = file->private_data;

	return videobuf_mmap_mapper(&v4l2->vm.vb_vidq, vma);
}

/**************************************************************
 * 					common struct
 **************************************************************/ 
static struct file_operations xi_fops = {
	.owner			= THIS_MODULE,
	.open           = xi_open,
	.release        = xi_close,
	.read           = xi_read,
	.poll			= xi_poll,
	.ioctl          = video_ioctl2,
	.mmap           = xi_mmap,
	.llseek         = no_llseek,
};

#define V4L2_STD_PRIVATE_MASK (V4L2_STD_NTSC_M | \
						 V4L2_STD_NTSC_443 | \
						 V4L2_STD_PAL_D | \
						 V4L2_STD_PAL_M | \
						 V4L2_STD_PAL_Nc | \
						 V4L2_STD_PAL_60 | \
						 V4L2_STD_SECAM_B | \
						 V4L2_STD_SECAM_D)

static struct v4l2_tvnorm tvnorms[] = {
	{
		.name      = "NTSC-M",
		.id        = V4L2_STD_NTSC_M,
	}
};

static struct video_device xi_template = {
	.name				  = "xi100xe",
	.type				  = VID_TYPE_CAPTURE,
	.fops           	  = &xi_fops,
	.minor				  = -1,
	.vidioc_querycap      = vidioc_querycap,
	.vidioc_enum_fmt_cap  = vidioc_enum_fmt_vid_cap,
	.vidioc_g_fmt_cap     = vidioc_g_fmt_vid_cap,
	.vidioc_try_fmt_cap   = vidioc_try_fmt_vid_cap,
	.vidioc_s_fmt_cap     = vidioc_s_fmt_vid_cap,
	.vidioc_reqbufs       = vidioc_reqbufs,
	.vidioc_querybuf      = vidioc_querybuf,
	.vidioc_qbuf          = vidioc_qbuf,
	.vidioc_dqbuf         = vidioc_dqbuf,
	.vidioc_enum_input    = vidioc_enum_input,
	.vidioc_g_input       = vidioc_g_input,
	.vidioc_s_input       = vidioc_s_input,
	.vidioc_queryctrl     = vidioc_queryctrl,
	.vidioc_g_ctrl        = vidioc_g_ctrl,
	.vidioc_s_ctrl        = vidioc_s_ctrl,
	.vidioc_streamon      = vidioc_streamon,
	.vidioc_streamoff     = vidioc_streamoff,
	.tvnorms              = tvnorms,
	.tvnormsize           = ARRAY_SIZE(tvnorms),
};
/******************************************************************************
							module ops
******************************************************************************/
int v4l2_instance_create(void *dev,int inst)
{
	XIVdrv *vdrv;
	struct video_device *vfd;
	int i, ret = -ENOMEM;
	
	vdrv = kzalloc(sizeof(*vdrv), GFP_KERNEL);
	if (!vdrv)
		return -ENOMEM;

	mutex_init(&vdrv->mutex);

	vfd = video_device_alloc();
	if (!vfd)
		goto unreg_dev;

	*vfd = xi_template;

	ret = video_register_device(vfd, VFL_TYPE_GRABBER, -1);
	if (ret < 0)
		goto rel_vdev;

	video_set_drvdata(vfd, vdrv);

	vdrv->configure = kzalloc(sizeof(int) * ARRAY_SIZE(xi_qctrl), GFP_KERNEL);
	if(NULL ==  vdrv->configure)
		goto rel_vdev;
	 
	for (i = 0; i < ARRAY_SIZE(xi_qctrl); i++)
		vdrv->configure[i] = xi_qctrl[i].default_value;

	list_add_tail(&vdrv->vlist, XI100XE_GetV4l2ListHead(dev));

	vdrv->used = 0;
	vdrv->vfd = vfd;
	vdrv->channal = inst;
	vdrv->dev = dev;
	
	vdrv->romid = XI102XEHD_GetROMID(dev,&vdrv->rlg);
	
	vdrv->vm.channal = inst;
	vdrv->vm.dev = dev;
	vdrv->vm.framerate = 0;
	vdrv->vm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	spin_lock_init(&vdrv->vm.slock);
	INIT_LIST_HEAD(&vdrv->vm.actives);
	init_waitqueue_head(&vdrv->vm.wq);
	
	return 0;

rel_vdev:
	video_unregister_device(vfd);
	video_device_release(vfd);
unreg_dev:
	kfree(vdrv);
	return ret;
}

int xi_v4l2_release(void *dev)
{
	XIVdrv *vdrv;
	struct list_head *list;
	struct list_head *head;
	
	head = XI100XE_GetV4l2ListHead(dev);

	while (!list_empty(head)) {
		list = head->next;
		list_del(list);
		vdrv = list_entry(list, XIVdrv, vlist);
		
		video_unregister_device(vdrv->vfd);
		video_device_release(vdrv->vfd);
		
		kfree(vdrv->configure);
		kfree(vdrv);
	}

	return 0;
}

int xi_v4l2_init(void *dev)
{
	int ret = -ENOMEM, i;

	for (i = 0; i < XI102XE_VINDEX_MAX; i++) {
		ret = v4l2_instance_create(dev,i);
		if (ret) {
			break;
		}
	}
	printk(KERN_INFO "n_v4l2s counts = %d\n",i);
	return ret;
}
