#include <linux/hardirq.h>
#include "OSAbstract/PrivateType.h"
#include "v4l2_centos6.h"

static VFmt formats[] = {
	{
		.name     = "4:2:2, packed, UYVY",
		.fourcc   = V4L2_PIX_FMT_UYVY,
		.depth    = 16,
	},
	{
		.name     = "4:2:2, packed, YUYV",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.depth    = 16,
	},
	{
		.name     = "RGB24",
		.fourcc   = V4L2_PIX_FMT_BGR24,
		.depth    = 24,
	},
	{
		.name     = "RGB32",
		.fourcc   = V4L2_PIX_FMT_BGR32,
		.depth    = 32,
	},
};
/*************************************************************
							misc ops
****************************************************************/
static VFmt *get_format(struct v4l2_format *f)
{
	unsigned int k;
		
	for (k = 0; k < ARRAY_SIZE(formats); k++) {
		if (formats[k].fourcc == f->fmt.pix.pixelformat)
			return &formats[k];
	}
	return NULL;
} 

/**************************************************************
 * 					kthread ops
 **************************************************************/ 
 static void v4l2_sleep(Vmanufactory *vm,unsigned long jcount)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&vm->wq, &wait);

	if ((!kthread_should_stop()) && (jcount > 0)) {
		schedule_timeout_interruptible(jcount);
	}
	remove_wait_queue(&vm->wq, &wait);
}

static void v4l2_framerate_adjust(Vmanufactory *vm)
{
	unsigned long jnext;
	
	if(!vm->framerate){
		vm->framerate = 25;
		return ;
	}
	
	if((!vm->prevrate) || (vm->prevrate != vm->framerate)){
		vm->prevrate = vm->framerate;
		vm->jbefore = jiffies;
		vm->jdelay = HZ/vm->framerate;
		return ;
	}
	
	jnext = vm->jbefore + vm->jdelay;
	if(time_is_after_jiffies(jnext)) {
		v4l2_sleep(vm,jnext - jiffies);
		vm->jbefore = jnext;
	} else {
		vm->jbefore = jiffies;
	}
}

static void v4l2_frame_syscby(Vmanufactory *vm, FrameSyscBy by)
{
	switch(by){
	case FRAME_SYSC_BY_IRQ:
		XI102XEHD_WaitForNewImage(vm->dev, vm->channal, HZ/10);
		break;
	default:
		v4l2_framerate_adjust(vm);
	}
}

static void V4l2_ProcessOneFrame(Vmanufactory *vm)
{
	unsigned long flags = 0;
	struct videobuf_buffer *vb;
	
	spin_lock_irqsave(&vm->slock, flags);
	if (list_empty(&vm->actives)) {
		spin_unlock_irqrestore(&vm->slock, flags);
		return;
	}

	vb = list_entry(vm->actives.next, struct videobuf_buffer, queue);
	if (!waitqueue_active(&vb->done)) {
		spin_unlock_irqrestore(&vm->slock, flags);
		return;
	}
	
	list_del(&vb->queue);
	spin_unlock_irqrestore(&vm->slock, flags);
	
	if(XI102XEHD_ProcessOneFrame(vm->dev, vm->channal, videobuf_to_vmalloc(vb))){
		vb->field_count++;
		vb->state = VIDEOBUF_DONE;
		wake_up(&vb->done);
	} else {
		vb->state = VIDEOBUF_ERROR;
		wake_up(&vb->done);
	}
}

/**********************************************************
							extern ops
***********************************************************/
static int Vmanufactory_Start(Vmanufactory *vm,int(*threadfn)(void *data))
{
	vm->field = V4L2_FIELD_ANY;
	vm->prevrate = 0;
	vm->field_count = 0;
	
	vm->pthread = kthread_run(threadfn, vm, "XI102XE_V4L2");
	if (IS_ERR(vm->pthread)) {
		return PTR_ERR(vm->pthread);
	}
	
	wake_up_interruptible(&vm->wq);

	return 0;
}

void Vmanufactory_Stop(Vmanufactory *vm)
{
	unsigned long flags = 0;
	
	if (vm->pthread) {
		kthread_stop(vm->pthread);
		vm->pthread = NULL;
	}
	
	spin_lock_irqsave(&vm->slock, flags);
	while(!list_empty(&vm->actives)) {
		struct videobuf_buffer *vb;
		vb = list_entry(vm->actives.next, struct videobuf_buffer, queue);
		list_del(&vb->queue);
		vb->state = VIDEOBUF_ERROR;
		wake_up(&vb->done);
	}
	spin_unlock_irqrestore(&vm->slock, flags);
}

int v4l2_kthread_machine(void *data)
{
	Vmanufactory *vm = data;
	
	set_freezable();
	while(true) {
		if (kthread_should_stop())
			break;
		
		V4l2_ProcessOneFrame(vm);
		v4l2_frame_syscby(vm,FRAME_SYSC_BY_IRQ);

		try_to_freeze();
	}
	
	return 0;
}

/******************************************************************************
							buffer ops
******************************************************************************/
static int v4l2_buffer_setup(struct videobuf_queue *vq, unsigned int *count, unsigned int *size)
{
	Vmanufactory *vm = vq->priv_data;
	*size = vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;

	if (0 == *count)
		*count = 32;

	while (*size * *count > VIDEO_BUF_MAX_M * 1024 * 1024)
		(*count)--;
	
	return 0;
}

static void v4l2_buffer_release(struct videobuf_queue *vq, struct videobuf_buffer *vb)
{
	if (in_interrupt())
		BUG();

	videobuf_waiton(vq,vb,0,0);
	videobuf_vmalloc_free(vb);

	vb->state = VIDEOBUF_NEEDS_INIT;
}

static int v4l2_buffer_prepare(struct videobuf_queue *vq, struct videobuf_buffer *vb,enum v4l2_field field)
{
	int rc;
	Vmanufactory *vm = vq->priv_data;

	if (vm->show.cxWidth  < 48 || vm->show.cxWidth  > 1920 ||
	    vm->show.cyHeight < 32 || vm->show.cyHeight > 1080)
		return -EINVAL;

	vb->size =  vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;
	if (0 != vb->baddr && vb->bsize < vb->size)
		return -EINVAL;
		
	vb->width  = vm->show.cxWidth;
	vb->height = vm->show.cyHeight;
	vb->bytesperline = vm->show.cxWidth * vm->show.depth >> 3;
	vb->field  = field;
	
	if (VIDEOBUF_NEEDS_INIT == vb->state) {
		rc = videobuf_iolock(vq, vb, NULL);
		if (rc < 0)
			goto fail;
	}

	vb->state = VIDEOBUF_PREPARED;
	return 0;

fail:
	v4l2_buffer_release(vq, vb);
	return rc;
}

static void v4l2_buffer_queue(struct videobuf_queue *vq, struct videobuf_buffer *vb)
{
	Vmanufactory *vm = vq->priv_data;
	vb->state = VIDEOBUF_QUEUED;
	
	list_add_tail(&vb->queue, &vm->actives);
}

static struct videobuf_queue_ops xi_video_qops = {
	.buf_setup      = v4l2_buffer_setup,
	.buf_prepare    = v4l2_buffer_prepare,
	.buf_queue      = v4l2_buffer_queue,
	.buf_release    = v4l2_buffer_release,
};

/******************************************************************************
							ioctl ops
******************************************************************************/
static int vidioc_querycap(struct file *file, void  *priv, struct v4l2_capability *cap)
{
	strcpy(cap->driver, "XI102XE");
	strcpy(cap->card, "XI102XE");
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE| V4L2_CAP_STREAMING | V4L2_CAP_READWRITE;

	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv, struct v4l2_fmtdesc *f)
{
	const VFmt *fmt;
	
	if (f->index >= ARRAY_SIZE(formats))
		return -EINVAL;

	fmt = &formats[f->index];

	strlcpy(f->description, fmt->name, sizeof(f->description));
	f->pixelformat = fmt->fourcc;

	return 0;
}


static int vidioc_g_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	f->fmt.pix.width        = vm->show.cxWidth;
	f->fmt.pix.height       = vm->show.cyHeight;
	f->fmt.pix.field        = vm->field;
	f->fmt.pix.pixelformat  = vm->show.fourcc;
	f->fmt.pix.bytesperline = (vm->show.cxWidth * vm->show.depth) >> 3;
	f->fmt.pix.sizeimage    = (vm->show.cxWidth * vm->show.cyHeight * vm->show.depth) >> 3;
	return 0;
}


static int vidioc_try_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	const VFmt *fmt;
	enum v4l2_field field;
	
	fmt = get_format(f);
	if (!fmt) {
		ZH_DEBUG(0,"Fourcc format (0x%08x) unknown.\n",f->fmt.pix.pixelformat);
		f->fmt.pix.pixelformat = formats[0].fourcc;
		//fmt = get_format(f);
		return -EINVAL;
	}

	field = f->fmt.pix.field;
	if (field == V4L2_FIELD_ANY) {
		field = V4L2_FIELD_NONE;
	} else if (V4L2_FIELD_NONE != field) {
		return -EINVAL;
	}
	
	f->fmt.pix.field = field;
	if (f->fmt.pix.height < 32) f->fmt.pix.height = 32;
	if (f->fmt.pix.height > MAX_HEIGHT) f->fmt.pix.height = MAX_HEIGHT;
	if (f->fmt.pix.width < 48) f->fmt.pix.width = 48;
	if (f->fmt.pix.width > MAX_WIDTH) f->fmt.pix.width = MAX_WIDTH;

	f->fmt.pix.width &= ~0x03;
	f->fmt.pix.bytesperline = (f->fmt.pix.width * fmt->depth) >> 3;
	f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;
		
	return 0;
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	int ret;
	VFmt *fmt = get_format(f);
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	struct videobuf_queue *q = &vm->vb_vidq;

	if (videobuf_queue_is_busy(q)) {
		return -EBUSY;
	}

	ret = vidioc_try_fmt_vid_cap(file, priv, f);
	if (ret < 0) {
		ZH_DEBUG(0,"Invalid fmt!\n");
		return ret;
	}
	
	if(NULL == fmt)
		fmt = &formats[0];
	
	mutex_lock(&q->vb_lock);
	vm->show.cxWidth 	= f->fmt.pix.width;
	vm->show.cyHeight 	= f->fmt.pix.height;
	vm->show.csOutput 	= V4L2FMT2CSOUT(fmt->fourcc);
	vm->show.fourcc 	= fmt->fourcc;
	vm->show.depth 		= fmt->depth;
	vm->show.bBottomUp  = true;
	vm->field 			= f->fmt.pix.field;
	mutex_unlock(&q->vb_lock);

	XI102XEHD_SetVideoShowInfo(
				vm->dev,
				vm->channal,
				true, 
				V4L2FMT2CSOUT(vm->show.fourcc),
				vm->show.cxWidth,
				vm->show.cyHeight
				);

	return 0;
}

static int vidioc_reqbufs(struct file *file, void *priv, struct v4l2_requestbuffers *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	return videobuf_reqbufs(&vm->vb_vidq, p);
}

static int vidioc_querybuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	return videobuf_querybuf(&vm->vb_vidq, p);
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return videobuf_qbuf(&vm->vb_vidq, p);
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
		
	return videobuf_dqbuf(&vm->vb_vidq, p, file->f_flags & O_NONBLOCK);
}

static int vidioc_streamon(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	if (vm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (i != vm->type)
		return -EINVAL;

	return videobuf_streamon(&vm->vb_vidq);
}

static int vidioc_streamoff(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	if (vm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (i != vm->type)
		return -EINVAL;

	return videobuf_streamoff(&vm->vb_vidq);
}

static int vidioc_enum_input(struct file *file, void *priv, struct v4l2_input *inp)
{
	XIVdrv *vdrv = video_drvdata(file);
	if (inp->index != 0)
		return -EINVAL;

	inp->type = V4L2_INPUT_TYPE_CAMERA;
	strcpy(inp->name, "XI102XE");
	inp->std = vdrv->vfd->tvnorms;

	return 0;
}

static int vidioc_g_parm(struct file *file, void *priv,  struct v4l2_streamparm *sp)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
		
	if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	sp->parm.capture.timeperframe.denominator = vm->framerate;
	sp->parm.capture.timeperframe.numerator = 1;
	return 0;
}

static int vidioc_s_parm(struct file *file, void *priv, struct v4l2_streamparm *sp)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	vm->framerate = sp->parm.capture.timeperframe.denominator/sp->parm.capture.timeperframe.numerator;
	return 0;
}

static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
	*i = 0;
	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
	if (i > 0)
		return -EINVAL;

	return (0);
}

/******************************************************************************
							fops ops
******************************************************************************/
static int xi_open(struct file *file)
{
	XIVdrv	*drv = video_drvdata(file);
	
	mutex_lock(&drv->mutex);
	if(drv->used) {
		mutex_unlock(&drv->mutex);
		return -EBUSY;
	} else {
		drv->used++;
		mutex_unlock(&drv->mutex);
	}
	
	drv->vm.framerate = 25;
	drv->vm.show.cxWidth = 1920;
	drv->vm.show.cyHeight = 1080;
	drv->vm.show.fourcc = V4L2_PIX_FMT_YUYV;
	drv->vm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	
	file->private_data = drv;

	videobuf_queue_vmalloc_init(
				&drv->vm.vb_vidq, 
				&xi_video_qops,
				NULL, 
				&drv->vm.slock,
				drv->vm.type, 
				V4L2_FIELD_NONE,
				sizeof(struct videobuf_buffer), 
				&drv->vm,
				NULL);

	return Vmanufactory_Start(&drv->vm,v4l2_kthread_machine);
}

static ssize_t xi_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
	XIVdrv *v4l2 = file->private_data;

	if (v4l2->vm.type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return videobuf_read_stream(&v4l2->vm.vb_vidq, data, count, ppos, 0,file->f_flags & O_NONBLOCK);
	}
	return 0;
}

static unsigned int xi_poll(struct file *file, struct poll_table_struct *wait)
{
	int ret;
	XIVdrv *v4l2 = file->private_data;
	struct videobuf_queue *q = &v4l2->vm.vb_vidq;
	
	if (V4L2_BUF_TYPE_VIDEO_CAPTURE != v4l2->vm.type)
		return POLLERR;
		
	ret =  videobuf_poll_stream(file, q, wait);
	return ret;
}

static int xi_close(struct file *file)
{
	XIVdrv *v4l2 = file->private_data;
	Vmanufactory_Stop(&v4l2->vm);

	videobuf_stop(&v4l2->vm.vb_vidq);
	videobuf_mmap_free(&v4l2->vm.vb_vidq);

	mutex_lock(&v4l2->mutex);
	v4l2->used = 0;
	mutex_unlock(&v4l2->mutex);

	return 0;
}

static int xi_mmap(struct file *file, struct vm_area_struct *vma)
{
	XIVdrv *v4l2 = file->private_data;

	return videobuf_mmap_mapper(&v4l2->vm.vb_vidq, vma);
}

/**************************************************************
 * 					common struct
 **************************************************************/ 
static const struct v4l2_file_operations xi_fops = {
	.owner			= THIS_MODULE,
	.open           = xi_open,
	.release        = xi_close,
	.read           = xi_read,
	.poll			= xi_poll,
	.unlocked_ioctl = video_ioctl2,
	.mmap           = xi_mmap,
};

static const struct v4l2_ioctl_ops xi_ioctl_ops = {
	.vidioc_querycap      		= vidioc_querycap,
	.vidioc_enum_fmt_vid_cap  	= vidioc_enum_fmt_vid_cap,
	.vidioc_g_fmt_vid_cap     	= vidioc_g_fmt_vid_cap,
	.vidioc_try_fmt_vid_cap   	= vidioc_try_fmt_vid_cap,
	.vidioc_s_fmt_vid_cap     	= vidioc_s_fmt_vid_cap,
	.vidioc_reqbufs       		= vidioc_reqbufs,
	.vidioc_querybuf      		= vidioc_querybuf,
	.vidioc_qbuf          		= vidioc_qbuf,
	.vidioc_dqbuf         		= vidioc_dqbuf,
	.vidioc_enum_input    		= vidioc_enum_input,
	.vidioc_g_input       		= vidioc_g_input,
	.vidioc_s_input       		= vidioc_s_input,
	.vidioc_streamon      		= vidioc_streamon,
	.vidioc_streamoff     		= vidioc_streamoff,
	
	/////////////////////////////////////////////////
	//.vidioc_queryctrl     	  	= vidioc_queryctrl,
	//.vidioc_g_ctrl        	  	= vidioc_g_ctrl,
	//.vidioc_s_ctrl        	  	= vidioc_s_ctrl,
	.vidioc_g_parm				= vidioc_g_parm,
	.vidioc_s_parm				= vidioc_s_parm,
};

#define V4L2_STD_P_MASK 		(V4L2_STD_PAL_D | V4L2_STD_PAL_M)
#define V4L2_STD_M_MASK 		(V4L2_STD_NTSC_M | \
								V4L2_STD_NTSC_443 | \
								V4L2_STD_PAL_D | \
								V4L2_STD_PAL_M | \
								V4L2_STD_PAL_Nc | \
								V4L2_STD_PAL_60 | \
								V4L2_STD_SECAM_B | \
								V4L2_STD_SECAM_D)

static struct video_device xi_template = {
	.name			= "XI102XE",
	.fops        	= &xi_fops,
	.ioctl_ops 		= &xi_ioctl_ops,
	.minor			= -1,
	.release		= video_device_release,

	.tvnorms  		= V4L2_STD_M_MASK,
	.current_norm 	= V4L2_STD_NTSC_M,
};
/******************************************************************************
							module ops
******************************************************************************/
static int video_nr = -1;
module_param(video_nr, uint, 0444);
MODULE_PARM_DESC(video_nr, "video iminor start number");

static int v4l2_instance_create(void *dev, int inst)
{
	int ret;
	XIVdrv *vdrv;
	struct video_device *vfd;
	
	vdrv = kzalloc(sizeof(*vdrv), GFP_KERNEL);
	if (!vdrv)
		return -ENOMEM;

	snprintf(vdrv->v4l2_dev.name, sizeof(vdrv->v4l2_dev.name),"%s-%03d", "XI102XE", inst);
	ret = v4l2_device_register(NULL, &vdrv->v4l2_dev);
	if (ret)
		goto v4l2_reg_fail;

	vdrv->vm.show.cxWidth  = 768;
	vdrv->vm.show.cyHeight = 576;
	vdrv->vm.show.csOutput = V4L2FMT2CSOUT(V4L2_PIX_FMT_UYVY);
	vdrv->vm.show.fourcc = V4L2_PIX_FMT_UYVY;
	vdrv->vm.show.depth = 16;
	vdrv->vm.show.bBottomUp = true;
	
	INIT_LIST_HEAD(&vdrv->vm.actives);
	init_waitqueue_head(&vdrv->vm.wq);
	vdrv->vm.channal = inst;
	vdrv->vm.dev = dev;
	vdrv->vm.jbefore = 0;
	vdrv->vm.jdelay = 0;
	vdrv->vm.prevrate = 0;
	vdrv->vm.framerate = 25;
	spin_lock_init(&vdrv->vm.slock);
	vdrv->vm.field = V4L2_FIELD_NONE;
	vdrv->vm.field_count = 0;
	vdrv->vm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	
	mutex_init(&vdrv->mutex);
	vdrv->used = 0;
	vdrv->channal = inst;
	vdrv->dev = dev;

	vfd = video_device_alloc();
	if (!vfd) {
		ret = -ENOMEM;
		goto vfd_alloc_fail;
	}
	*vfd = xi_template;
	vfd->debug = 0;
	
	vfd->v4l2_dev = &vdrv->v4l2_dev;
		
	ret = video_register_device(vfd, VFL_TYPE_GRABBER, -1);
	if (ret < 0)
		goto vfd_reg_fail;

	video_set_drvdata(vfd, vdrv);
	vdrv->vfd = vfd;

	list_add_tail(&vdrv->vlist, XI102XEHD_GetV4l2ListHead(dev));
	return 0;

vfd_reg_fail:
	video_device_release(vfd);
vfd_alloc_fail:
	v4l2_device_unregister(&vdrv->v4l2_dev);
v4l2_reg_fail:
	kfree(vdrv);
	return ret;
}

int xi_v4l2_release(void *dev)
{
	struct list_head *head = XI102XEHD_GetV4l2ListHead(dev);
	
	while (!list_empty(head)) {
		XIVdrv *vdrv = list_entry(head->next, XIVdrv, vlist);
		list_del(&vdrv->vlist);
		video_unregister_device(vdrv->vfd);
		v4l2_device_unregister(&vdrv->v4l2_dev);
		kfree(vdrv);
	}

	return 0;
}

int xi_v4l2_init(void *dev)
{
	int ret = -ENOMEM, i;

	for (i = 0; i < XI102XE_VINDEX_MAX; i++) {
		ret = v4l2_instance_create(dev,i);
		if (ret) 
			break;
	}
	ZH_DEBUG(1, "video counts = %d\n",i);
	return ret;
}
