
#include "OSAbstract/PrivateType.h"
#include "v4l2.h"

#define 	FPS_MAX 1000
static const struct v4l2_fract tpf_min = {
	.numerator = 1,
	.denominator = FPS_MAX
};

static const struct v4l2_fract tpf_max = {
	.numerator = FPS_MAX,
	.denominator = 1
};

static const struct v4l2_fract tpf_default = {
	.numerator = 1001,
	.denominator = 30000
};

#define frames_to_ms(vm, frames) 	((frames * vm->tpf.numerator * 1000) / vm->tpf.denominator)

static VFmt formats[] = {
	{
		.name     = "4:2:2, packed, UYVY",
		.fourcc   = V4L2_PIX_FMT_UYVY,
		.depth    = 16,
	},
		{
		.name     = "4:2:2, packed, YUYV",
		.fourcc   = V4L2_PIX_FMT_YUYV,
		.depth    = 16,
	},
	{
		.name     = "RGB24",
		.fourcc   = V4L2_PIX_FMT_BGR24,
		.depth    = 24,
	},
	{
		.name     = "RGB32",
		.fourcc   = V4L2_PIX_FMT_BGR32,
		.depth    = 32,
	},
};
/*************************************************************
							misc ops
****************************************************************/
static VFmt *get_format(struct v4l2_format *f)
{
	unsigned int k;
		
	for (k = 0; k < ARRAY_SIZE(formats); k++) {
		if (formats[k].fourcc == f->fmt.pix.pixelformat)
			return &formats[k];
	}
	return NULL;
} 

/**************************************************************
 * 					kthread ops
 **************************************************************/ 

 
 static void v4l2_sleep(Vmanufactory *vm,unsigned long jcount)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&vm->wq, &wait);

	if ((!kthread_should_stop()) && (jcount > 0)) {
		schedule_timeout_interruptible(jcount);
	}
	remove_wait_queue(&vm->wq, &wait);
}

static void v4l2_framerate_adjust(Vmanufactory *vm)
{
	unsigned long jdelay, anchor, flags = 0;

	spin_lock_irqsave(&vm->tpf_slock, flags);
	if((vm->tpf.denominator * vm->tpf_before.numerator) 
			!= (vm->tpf_before.denominator * vm->tpf.numerator)) {
		vm->tpf_before = vm->tpf;
		vm->frames = 0;
		vm->jiffies = jiffies;
	}
	spin_unlock_irqrestore(&vm->tpf_slock, flags);
	
	vm->frames++;
	jdelay = HZ * vm->frames * vm->tpf.numerator / vm->tpf.denominator;
	anchor = vm->jiffies + jdelay;
	
	if(time_is_after_jiffies(anchor)) {
		v4l2_sleep(vm, anchor - jiffies);
	}
}

static void v4l2_wait_frame(Vmanufactory *vm)
{
	if(atomic_read(&vm->byirq)) {
		XI200DESDIDevice_WaitForNewImage(vm->dev, vm->channal, HZ/10);
	} else {
		v4l2_framerate_adjust(vm);
	}
}

static void V4l2_ProcessOneFrame(Vmanufactory *vm)
{
	unsigned long flags = 0;
	VBuf *vbuf;
	
	spin_lock_irqsave(&vm->slock, flags);
	if (list_empty(&vm->actives)) {
		spin_unlock_irqrestore(&vm->slock, flags);
		return;
	}

	vbuf = list_entry(vm->actives.next, VBuf, buflist);
	list_del(&vbuf->buflist);
	spin_unlock_irqrestore(&vm->slock, flags);
	
	if(0 == XI200DESDIDevice_ProcessOneFrame(vm->dev, vm->channal, vb2_plane_vaddr(&vbuf->vb,0))){
		do_gettimeofday(&vbuf->vb.v4l2_buf.timestamp);
		vbuf->vb.v4l2_buf.field = vm->field;
		vbuf->vb.v4l2_buf.sequence = vm->field_count++ ;
		vb2_buffer_done(&vbuf->vb,VB2_BUF_STATE_DONE);
	} else {
		do_gettimeofday(&vbuf->vb.v4l2_buf.timestamp);
		vb2_buffer_done(&vbuf->vb,VB2_BUF_STATE_ERROR);
	}
}

/**********************************************************
							extern ops
***********************************************************/
static int Vmanufactory_Start(Vmanufactory *vm,int(*threadfn)(void *data))
{
	XIVdrv *vdrv = container_of(vm, XIVdrv, vm);
	
	if(atomic_inc_return(&vdrv->used) > 1) {
		atomic_dec_return(&vdrv->used);
		return -EBUSY;
	}
	
	vm->pthread = kthread_run(threadfn, vm, "XI200DESDI_V4L2");
	if (IS_ERR(vm->pthread)) {
		return PTR_ERR(vm->pthread);
	}
	
	wake_up_interruptible(&vm->wq);

	return 0;
}

void Vmanufactory_Stop(Vmanufactory *vm)
{
	unsigned long flags = 0;
	XIVdrv *vdrv = container_of(vm, XIVdrv, vm);
	
	if (vm->pthread) {
		kthread_stop(vm->pthread);
		vm->pthread = NULL;
	}
	
	spin_lock_irqsave(&vm->slock, flags);
	while(!list_empty(&vm->actives)) {
		VBuf *vbuf = list_entry(vm->actives.next,
				VBuf, buflist);
		list_del_init(&vbuf->buflist);
		vb2_buffer_done(&vbuf->vb,VB2_BUF_STATE_ERROR);
	}
	spin_unlock_irqrestore(&vm->slock, flags);
	
	atomic_dec_return(&vdrv->used);
}

int v4l2_kthread_machine(void *data)
{
	Vmanufactory *vm = data;
	
	ZH_DEBUG(1,"Thread started\n");
	
	vm->field = V4L2_FIELD_ANY;
	vm->field_count = 0;
	vm->frames = 0;
	vm->jiffies = jiffies;	

	set_freezable();
	while(true) {
		if (kthread_should_stop())
			break;
		
		v4l2_wait_frame(vm);
		V4l2_ProcessOneFrame(vm);
		//v4l2_frame_syscby(vm,FRAME_SYSC_BY_IRQ);

		try_to_freeze();
	}
	ZH_DEBUG(1,"Thread stoped\n");
	
	return 0;
}
/*********************************************************
							buffer ops
**********************************************************/
static int queue_setup(
				struct vb2_queue *vq, 
					#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,2,0)
				const struct v4l2_format *fmt,
					#endif 
				unsigned int *nbuffers, 
				unsigned int *nplanes,
					#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,2,0)
				unsigned int sizes[], 
					#else
				unsigned long sizes[], 
					#endif
				void *alloc_ctxs[]
				)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	unsigned long size = 	vm->show.cxWidth * \
							vm->show.cyHeight * \
							vm->show.depth >> 3;
	
	if (0 == *nbuffers)
		*nbuffers = 32;
		
	if (size == 0)
		return -EINVAL;

	while (size * *nbuffers > VIDEO_BUF_MAX_M * 1024 * 1024)
		(*nbuffers)--;
		
	*nplanes = 1;
	sizes[0] = size;
	return 0;
}


static int buffer_init(struct vb2_buffer *vb)
{
	return 0;
}

static int buffer_prepare(struct vb2_buffer *vb)
{
	unsigned long size;
	Vmanufactory *vm = vb2_get_drv_priv(vb->vb2_queue);

	if (vm->show.cxWidth  < 48 || vm->show.cxWidth  > MAX_WIDTH ||
	    vm->show.cyHeight < 32 || vm->show.cyHeight > MAX_HEIGHT)
		return -EINVAL;

	size = vm->show.cxWidth * vm->show.cyHeight * vm->show.depth >> 3;
	if (vb2_plane_size(vb, 0) < size) {
		return -EINVAL;
	}

	vb2_set_plane_payload(vb, 0, size);
	
	return 0;
}

static int buffer_finish(struct vb2_buffer *vb)
{
	return 0;
}

static void buffer_cleanup(struct vb2_buffer *vb)
{
	return ;
}

static void buffer_queue(struct vb2_buffer *vb)
{
	Vmanufactory *vm = vb2_get_drv_priv(vb->vb2_queue);
	VBuf *vbuf = container_of(vb, VBuf, vb);
	unsigned long flags = 0;

	spin_lock_irqsave(&vm->slock, flags);
	list_add_tail(&vbuf->buflist, &vm->actives);
	spin_unlock_irqrestore(&vm->slock, flags);
}

static int start_streaming(
					struct vb2_queue *vq 
						#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,2,0)
					, unsigned int count
						#endif
					)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	return Vmanufactory_Start(vm,v4l2_kthread_machine);
}

static int stop_streaming(struct vb2_queue *vq)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	Vmanufactory_Stop(vm);
	return 0;
}

static void xi_lock(struct vb2_queue *vq)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	XIVdrv *drv = container_of(vm, XIVdrv, vm);
	mutex_lock(&drv->mutex);
}

static void xi_unlock(struct vb2_queue *vq)
{
	Vmanufactory *vm = vb2_get_drv_priv(vq);
	XIVdrv *drv = container_of(vm, XIVdrv, vm);
	mutex_unlock(&drv->mutex);
}

static struct vb2_ops xi_video_qops = {
	.queue_setup		= queue_setup,
	.buf_init			= buffer_init,
	.buf_prepare		= buffer_prepare,
	.buf_finish			= buffer_finish,
	.buf_cleanup		= buffer_cleanup,
	.buf_queue			= buffer_queue,
	.start_streaming	= start_streaming,
	.stop_streaming		= stop_streaming,
	.wait_prepare		= xi_unlock,
	.wait_finish		= xi_lock,
};

/******************************************************************************
							ioctl ops
******************************************************************************/
static int vidioc_querycap(struct file *file, void  *priv, struct v4l2_capability *cap)
{
	strcpy(cap->driver, "XI200DESDI");
	strcpy(cap->card, "XI200DESDI");
	cap->device_caps = V4L2_CAP_VIDEO_CAPTURE|
						V4L2_CAP_STREAMING   |
						V4L2_CAP_READWRITE;
	cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;

	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv, struct v4l2_fmtdesc *f)
{
	VFmt *fmt;
	
	if (f->index >= ARRAY_SIZE(formats))
		return -EINVAL;

	fmt = &formats[f->index];

	strlcpy(f->description, fmt->name, sizeof(f->description));
	f->pixelformat = fmt->fourcc;

	return 0;
}

static int vidioc_g_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	long lImageFormatVersion;
	bool bInterlaced, bPrescaled, bHasSignal;
	int cx, cy, cxScaled;
	VINEX_FORMAT format; 

	XI200DESDIDevice_GetInputSignal(
			vm->dev,
			vm->channal,
			&lImageFormatVersion, 
			&bInterlaced, 
			&cx, 
			&cy, 
			&format, 
			&bPrescaled, 
			&cxScaled, 
			&bHasSignal
			);

	if(bHasSignal) {
		vm->show.cxWidth 	= cx;
		vm->show.cyHeight 	= cy;
		vm->show.csOutput 	= VinexFormat2ColorSpace(format);
		vm->show.fourcc 	= VinexFormat2Fourcc(format);
		vm->show.depth 		= GetColorSpaceBpp(vm->show.csOutput);
	}
		
	f->fmt.pix.width        = vm->show.cxWidth;
	f->fmt.pix.height       = vm->show.cyHeight;
	f->fmt.pix.field        = vm->field;
	f->fmt.pix.pixelformat  = vm->show.fourcc;
	f->fmt.pix.bytesperline = (vm->show.cxWidth * 
								vm->show.cyHeight) >> 3;
	f->fmt.pix.sizeimage    = (vm->show.cxWidth * 
								vm->show.cyHeight * 
								vm->show.depth) >> 3;
	return 0;
}

static int vidioc_try_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	const VFmt *fmt;
	enum v4l2_field field;
	
	fmt = get_format(f);
	if (!fmt) {
		ZH_DEBUG(1,"Fourcc format (0x%08x) unknown.\n",f->fmt.pix.pixelformat);
		f->fmt.pix.pixelformat = formats[0].fourcc;
		fmt = get_format(f);
	}

	field = f->fmt.pix.field;
	if (field == V4L2_FIELD_ANY) {
		field = V4L2_FIELD_NONE;
	} else if (V4L2_FIELD_NONE != field) {
		return -EINVAL;
	}
	
	f->fmt.pix.field = field;
	v4l_bound_align_image(
					&f->fmt.pix.width, 
					48, 
					MAX_WIDTH, 
					2,
					&f->fmt.pix.height, 
					32, 
					MAX_HEIGHT, 
					0, 
					0
					);

	f->fmt.pix.bytesperline = (f->fmt.pix.width * fmt->depth) >> 3;
	f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;
		
	return 0;
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv, struct v4l2_format *f)
{
	int ret;
	VFmt *fmt = get_format(f);
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	struct vb2_queue *q = &vm->vb_vidq;	

	if (vb2_is_streaming(q)) {
		return -EBUSY;
	}

	ret = vidioc_try_fmt_vid_cap(file, priv, f);
	if (ret < 0) {
		ZH_DEBUG(1,"Invalid fmt!\n");
		return ret;
	}
	
	if(NULL == fmt)
		fmt = &formats[0];
			
	vm->show.cxWidth 	= f->fmt.pix.width;
	vm->show.cyHeight 	= f->fmt.pix.height;
	vm->show.csOutput 	= V4L2FMT2CSOUT(fmt->fourcc);
	vm->show.fourcc 	= fmt->fourcc;
	vm->show.depth 		= fmt->depth;
	vm->show.bBottomUp  = true;
	vm->field 			= f->fmt.pix.field;

	XI200DESDIDevice_SetVideoShowInfo(
				vm->dev,
				vm->channal,
				true, 
				V4L2FMT2CSOUT(vm->show.fourcc),
				vm->show.cxWidth,
				vm->show.cyHeight
				);

	return 0;
}

static int vidioc_enum_frameintervals(struct file *file, void *priv, struct v4l2_frmivalenum *fival)
{
	if (fival->index)
		return -EINVAL;

	fival->type = V4L2_FRMIVAL_TYPE_CONTINUOUS;

	fival->stepwise.min  = tpf_min;
	fival->stepwise.max  = tpf_max;
	fival->stepwise.step = (struct v4l2_fract) {1, 1};

	return 0;
}

static int vidioc_g_parm(struct file *file, void *priv, struct v4l2_streamparm *parm)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	if (parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	parm->parm.capture.capability   = V4L2_CAP_TIMEPERFRAME;
	parm->parm.capture.timeperframe = vm->tpf;
	parm->parm.capture.readbuffers  = 1;
	return 0;
}

#define FRACT_CMP(a, OP, b)	\
	((u64)(a).numerator * (b).denominator  OP  (u64)(b).numerator * (a).denominator)

static int vidioc_s_parm(struct file *file, void *priv, struct v4l2_streamparm *parm)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	struct v4l2_fract tpf;

	if (parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	tpf = parm->parm.capture.timeperframe;

	tpf = tpf.denominator ? tpf : tpf_default;
	tpf = FRACT_CMP(tpf, <, tpf_min) ? tpf_min : tpf;
	tpf = FRACT_CMP(tpf, >, tpf_max) ? tpf_max : tpf;

	vm->tpf = tpf;
	parm->parm.capture.timeperframe = tpf;
	parm->parm.capture.readbuffers  = 1;

	return 0;
}

static int vidioc_reqbufs(struct file *file, void *priv, struct v4l2_requestbuffers *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	return vb2_reqbufs(&vm->vb_vidq, p);
}

static int vidioc_querybuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;

	return vb2_querybuf(&vm->vb_vidq, p);
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return vb2_qbuf(&vm->vb_vidq, p);
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *p)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
		
	return vb2_dqbuf(&vm->vb_vidq, p, file->f_flags & O_NONBLOCK);
}

static int vidioc_streamon(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return vb2_streamon(&vm->vb_vidq, i);
}

static int vidioc_streamoff(struct file *file, void *priv, enum v4l2_buf_type i)
{
	XIVdrv *drv = video_drvdata(file);
	Vmanufactory *vm = &drv->vm;
	
	return vb2_streamoff(&vm->vb_vidq, i);
}

static int vidioc_enum_input(struct file *file, void *priv, struct v4l2_input *inp)
{
	XIVdrv *vdrv = video_drvdata(file);
	if (inp->index != 0)
		return -EINVAL;

	inp->type = V4L2_INPUT_TYPE_CAMERA;
	sprintf(inp->name, "XI200DESDI: %u", vdrv->channal);
	inp->std = vdrv->vfd->tvnorms;

	return 0;
}

static int vidioc_g_input(struct file *file, void *priv, unsigned int *i)
{
	*i = 0;
	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int i)
{
	if (i > 0)
		return -EINVAL;

	return (0);
}

//get signal info 
static long vidioc_default(struct file *file, void *fh,
					bool valid_prio, unsigned int cmd, void *arg)
{
	XIVdrv *drv = video_drvdata(file);
	
	switch (cmd) {
	case XI_SIGNAL_INFO:
		return XI200DESDIDevice_GetSignalInfo(drv->dev, drv->channal,(VIDEO_SIGNAL_INFO *) arg);
		
	default:
		return -ENOTTY;
	}
}

static const struct v4l2_ioctl_ops xi_ioctl_ops = {
	.vidioc_querycap      		= vidioc_querycap,
	.vidioc_enum_fmt_vid_cap  	= vidioc_enum_fmt_vid_cap,
	.vidioc_g_fmt_vid_cap     	= vidioc_g_fmt_vid_cap,
	.vidioc_try_fmt_vid_cap   	= vidioc_try_fmt_vid_cap,
	.vidioc_s_fmt_vid_cap     	= vidioc_s_fmt_vid_cap,
	.vidioc_reqbufs       		= vidioc_reqbufs,
	.vidioc_querybuf      		= vidioc_querybuf,
	.vidioc_qbuf          		= vidioc_qbuf,
	.vidioc_dqbuf         		= vidioc_dqbuf,
	.vidioc_enum_frameintervals = vidioc_enum_frameintervals,
	.vidioc_g_parm				= vidioc_g_parm,
	.vidioc_s_parm				= vidioc_s_parm,
	.vidioc_enum_input    		= vidioc_enum_input,
	.vidioc_g_input       		= vidioc_g_input,
	.vidioc_s_input       		= vidioc_s_input,
	.vidioc_streamon      		= vidioc_streamon,
	.vidioc_streamoff     		= vidioc_streamoff,
	.vidioc_default				= vidioc_default,
};
/******************************************************************************
							fs ops
******************************************************************************/
static ssize_t xi_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
	XIVdrv *drv = video_drvdata(file);

	return vb2_read(&drv->vm.vb_vidq, data, count, ppos, file->f_flags & O_NONBLOCK);
}

static unsigned int xi_poll(struct file *file, struct poll_table_struct *wait)
{
	XIVdrv *drv = video_drvdata(file);
	struct vb2_queue *q = &drv->vm.vb_vidq;
	
	return vb2_poll(q, file, wait);
}

static int xi_close(struct file *file)
{
	XIVdrv *drv = video_drvdata(file);

	if (v4l2_fh_is_singular_file(file))
		vb2_queue_release(&drv->vm.vb_vidq);
		
	return v4l2_fh_release(file);
}

static int xi_mmap(struct file *file, struct vm_area_struct *vma)
{
	XIVdrv *drv = video_drvdata(file);
	
	return vb2_mmap(&drv->vm.vb_vidq, vma);
}

static const struct v4l2_file_operations xi_fops = {
	.owner			= THIS_MODULE,
	.open           = v4l2_fh_open,
	.release        = xi_close,
	.read           = xi_read,
	.poll			= xi_poll,
	.unlocked_ioctl = video_ioctl2,
	.mmap           = xi_mmap,
};

/******************************************************************************
							dev struct
******************************************************************************/
#define V4L2_STD_P_MASK 		(V4L2_STD_PAL_D | V4L2_STD_PAL_M)
#define V4L2_STD_M_MASK 		(V4L2_STD_NTSC_M | \
								V4L2_STD_NTSC_443 | \
								V4L2_STD_PAL_D | \
								V4L2_STD_PAL_M | \
								V4L2_STD_PAL_Nc | \
								V4L2_STD_PAL_60 | \
								V4L2_STD_SECAM_B | \
								V4L2_STD_SECAM_D)

static struct video_device xi_template = {
	.name			= "XI200DESDI",
	.fops        	= &xi_fops,
	.ioctl_ops 		= &xi_ioctl_ops,
	.release		= video_device_release,

	.tvnorms  		= V4L2_STD_M_MASK,
	//.current_norm 	= V4L2_STD_NTSC_M,
};
/******************************************************************************
							ioctl ops
******************************************************************************/
#define XI_CID_CUSTOM_BASE	(V4L2_CID_USER_BASE | 0xf900)

static int xi_s_ctrl(struct v4l2_ctrl *ctrl);
static const struct v4l2_ctrl_ops xi_ctrl_ops = {
	.s_ctrl 		 = xi_s_ctrl,
};

//Capture By Irq
static const struct v4l2_ctrl_config xi_ctrl_cbi = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_BASE + 0,
	.name = "Capture By Irq",
	.type = V4L2_CTRL_TYPE_BOOLEAN,
	.min = 0,
	.max = 1,
	.step = 1,
	.def = 1,
};

//set deinterlace method
static const char * const xi_ctrl_deinterlace[] = {
	"DEINTERLACE_NONE",
	"DEINTERLACE_BLEND",
	NULL,
};

static const struct v4l2_ctrl_config xi_ctrl_deinterlace_menu = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_BASE + 5,
	.name = "Deinterlace",
	.type = V4L2_CTRL_TYPE_MENU,
	.min = 0,
	.max = 1,
	.def = 1,
	//.menu_skip_mask = 0x04,
	.qmenu = xi_ctrl_deinterlace,
};

//set scale mode 
static const char * const xi_ctrl_scale[] = {
	"SCALEP_TYPE_SQEEZE",
	"SCALEP_TYPE_LETTER_BOX",
	"SCALEP_TYPE_EDGE_CROP",
	NULL,
};

static const struct v4l2_ctrl_config xi_ctrl_scale_menu = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_BASE + 6,
	.name = "Scale",
	.type = V4L2_CTRL_TYPE_MENU,
	.min = 0,
	.max = 2,
	.def = 1,
	//.menu_skip_mask = 0x04,
	.qmenu = xi_ctrl_scale,
};

//set phase
static const struct v4l2_ctrl_config xi_ctrl_phase = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_BASE + 7,
	.name = "Phase",
	.type = V4L2_CTRL_TYPE_INTEGER,
	.min = -128,
	.max = 127,
	.def = 0,
};

//set Video extern format
//set scale mode 
static const char * const xi_ctrl_ext_format[] = {
	"VINEX_FORMAT_RGB32",
	"VINEX_FORMAT_RGB24",
	"VINEX_FORMAT_AYUV",
	"VINEX_FORMAT_YUYV",
	"VINEX_FORMAT_YVYU",
	"VINEX_FORMAT_UYVY",	
	"VINEX_FORMAT_VYUY",
	"VINEX_FORMAT_GRAY8",
	NULL,
};

static const struct v4l2_ctrl_config xi_ctrl_eformat_menu = {
	.ops = &xi_ctrl_ops,
	.id = XI_CID_CUSTOM_BASE + 8,
	.name = "Extern Format",
	.type = V4L2_CTRL_TYPE_MENU,
	.min = 0,
	.max = 7,
	.def = 4,
	.menu_skip_mask = (0x1 << 2) | (0x1 << 4) | (0x1 << 6) | (0x1 << 7),
	.qmenu = xi_ctrl_ext_format,
};

/////
static int xi_s_ctrl(struct v4l2_ctrl *ctrl)
{
	long done;
	XIVdrv *drv = container_of(ctrl->handler, XIVdrv, ctrl_handler);
	
	switch (ctrl->id) {
		case V4L2_CID_BRIGHTNESS:
			done = XI200DESDIDevice_SetBrightness(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;
			break;
			
		case V4L2_CID_CONTRAST:
			done = XI200DESDIDevice_SetContrast(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;			
			break;	
			
		case V4L2_CID_SATURATION:
			done = XI200DESDIDevice_SetSaturation(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;	
			break;
			
		case V4L2_CID_HUE:
			done = XI200DESDIDevice_SetHue(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;	
			break;
			
		case V4L2_CID_SHARPNESS:
			done = XI200DESDIDevice_SetSharpness(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;	
			break;
			
		case V4L2_CID_HFLIP:
			done = XI200DESDIDevice_SetHorizontalFlip(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;	
			break;
			
		case V4L2_CID_VFLIP:
			done = XI200DESDIDevice_SetVerticalFlip(drv->dev, drv->channal, ctrl->val);
			if(done != ctrl->val)
				return -EPERM;	
				
		case XI_CID_CUSTOM_BASE + 0:
			atomic_set(&drv->vm.byirq,ctrl->val);			
			break;
			
		case XI_CID_CUSTOM_BASE + 5:
			return XI200DESDIDevice_SetDeinterlaceMethod(drv->dev, drv->channal, ctrl->val);
			
		case XI_CID_CUSTOM_BASE + 6:
			return XI200DESDIDevice_SetScaleMode(drv->dev, drv->channal, ctrl->val);			
			
		case XI_CID_CUSTOM_BASE + 7:
			return XI200DESDIDevice_SetPhase(drv->dev, drv->channal, ctrl->val);
			
		case XI_CID_CUSTOM_BASE + 8:
			return XI200DESDIDevice_SetFormat(drv->dev, drv->channal, ctrl->val);
			
		default:
			break;
	}

	return 0;
}

/******************************************************************************
							module ops
******************************************************************************/
static int __init v4l2_instance_create(void *dev, int inst)
{
	int ret;
	XIVdrv *vdrv;
	struct video_device *vfd;
	struct v4l2_ctrl_handler *hdl;
	struct vb2_queue *q;

	vdrv = kzalloc(sizeof(*vdrv), GFP_KERNEL);
	if (!vdrv)
		return -ENOMEM;

	snprintf(vdrv->v4l2_dev.name, sizeof(vdrv->v4l2_dev.name),"%s-%03d", "XI200DESDI", inst);
	ret = v4l2_device_register(NULL, &vdrv->v4l2_dev);
	if (ret)
		goto v4l2_reg_fail;

	vdrv->vm.show.cxWidth  = 768;
	vdrv->vm.show.cyHeight = 576;
	vdrv->vm.show.csOutput = V4L2FMT2CSOUT(V4L2_PIX_FMT_UYVY);
	vdrv->vm.show.fourcc = V4L2_PIX_FMT_UYVY;
	vdrv->vm.show.depth = 16;
	vdrv->vm.show.bBottomUp = true;
	
	INIT_LIST_HEAD(&vdrv->vm.actives);
	init_waitqueue_head(&vdrv->vm.wq);
	vdrv->vm.channal = inst;
	vdrv->vm.dev = dev;
	
	spin_lock_init(&vdrv->vm.tpf_slock);
	vdrv->vm.tpf = tpf_default;
	vdrv->vm.tpf_before = tpf_default;
	vdrv->vm.frames = 0;
	vdrv->vm.jiffies = jiffies;	
	atomic_set(&vdrv->vm.byirq,0);
	
	spin_lock_init(&vdrv->vm.slock);
	vdrv->vm.field = V4L2_FIELD_NONE;
	vdrv->vm.field_count = 0;
	mutex_init(&vdrv->mutex);
	atomic_set(&vdrv->used,0);
	vdrv->channal = inst;
	vdrv->dev = dev;
	
	//add ctrls
	hdl = &vdrv->ctrl_handler;
	v4l2_ctrl_handler_init(hdl, 17);
	ZH_DEBUG(1,"v4l2_ctrl_handler_init ----->!\n");
	
	vdrv->contrast = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops, V4L2_CID_CONTRAST, 0, 255, 1, 127);//need check
	vdrv->brightness = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops,	V4L2_CID_BRIGHTNESS, 0, 255, 1, 10);//need check
	vdrv->hue = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops, V4L2_CID_HUE, -128, 127, 1, 0);	//need check
	vdrv->saturation = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops,	V4L2_CID_SATURATION, 0, 255, 1, 164);//need check
	vdrv->sharpness = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops, V4L2_CID_SHARPNESS, -128, 127, 1, 0);//need check
	
	vdrv->hflip = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops, V4L2_CID_HFLIP, 0, 1, 1, 0);
	vdrv->vflip = v4l2_ctrl_new_std(hdl, &xi_ctrl_ops, V4L2_CID_VFLIP, 0, 1, 1, 0);		
	vdrv->sysc_byirq = v4l2_ctrl_new_custom(hdl, &xi_ctrl_cbi, NULL);	
	vdrv->deinterlace = v4l2_ctrl_new_custom(hdl, &xi_ctrl_deinterlace_menu, NULL);
	vdrv->scale = v4l2_ctrl_new_custom(hdl, &xi_ctrl_scale_menu, NULL);
	//vdrv->eformat = v4l2_ctrl_new_custom(hdl, &xi_ctrl_eformat_menu, NULL);
	ZH_DEBUG(1,"<-----v4l2_ctrl_handler_init !\n");
	
	if (hdl->error) {
		ret = hdl->error;
		goto vdrv_hdl_fail;
	}
	vdrv->v4l2_dev.ctrl_handler = hdl;
	v4l2_ctrl_handler_setup(hdl);
	//ctrl ops end 
	
	q = &vdrv->vm.vb_vidq;
	memset(q, 0, sizeof(vdrv->vm.vb_vidq));
	q->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	q->io_modes = VB2_MMAP | VB2_USERPTR | VB2_READ;
	q->drv_priv = &vdrv->vm;
	q->buf_struct_size = sizeof(VBuf);
	q->ops = &xi_video_qops;
	q->mem_ops = &vb2_vmalloc_memops;
	vb2_queue_init(q);

	vfd = video_device_alloc();
	if (!vfd) {
		ret = -ENOMEM;
		goto vfd_alloc_fail;
	}
	*vfd = xi_template;
	vfd->debug = 0;
	vfd->v4l2_dev = &vdrv->v4l2_dev;
	set_bit(V4L2_FL_USES_V4L2_FH, &vfd->flags);
	vfd->lock = &vdrv->mutex;
	ret = video_register_device(vfd, VFL_TYPE_GRABBER, -1);
	if (ret < 0)
		goto vfd_reg_fail;

	video_set_drvdata(vfd, vdrv);
	vdrv->vfd = vfd;
	
	list_add_tail(&vdrv->vlist, XI200DESDIDevice_GetV4l2ListHead(dev));
	return 0;
	
vfd_reg_fail:
	video_device_release(vfd);
vfd_alloc_fail:
	v4l2_device_unregister(&vdrv->v4l2_dev);
vdrv_hdl_fail:
	v4l2_ctrl_handler_free(hdl);
v4l2_reg_fail:
	kfree(vdrv);
	return ret;
}

int xi_v4l2_release(void *dev)
{
	struct list_head *head = XI200DESDIDevice_GetV4l2ListHead(dev);
	
	while (!list_empty(head)) {
		XIVdrv *vdrv = list_entry(head->next, XIVdrv, vlist);
		list_del(&vdrv->vlist);
		
		video_unregister_device(vdrv->vfd);
		v4l2_device_unregister(&vdrv->v4l2_dev);
		v4l2_ctrl_handler_free(&vdrv->ctrl_handler);
		
		kfree(vdrv);
	}

	return 0;
}

int xi_v4l2_init(void *dev)
{
	int ret = -ENOMEM, i;

	for (i = 0; i < XI200DE_VIDEO_DEVICE_COUNT; i++) {
		ret = v4l2_instance_create(dev,i);
		if (ret) 
			break;
	}
	ZH_DEBUG(1, "video counts = %d\n",i);
	return ret;
}
