#include <linux/module.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/mutex.h>
#include <linux/videodev2.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22)
#include <linux/freezer.h>
#endif

#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>

#include <asm/atomic.h>
#include <asm/i387.h>
#include "OSAbstract/OSAbstract.h"

////////////////////////////////////////////////////////////////////////////////
// wait queue
struct _OS_WaitQueueHeadT{
	wait_queue_head_t t;
};

OS_WaitQueueHeadT * OS_WaitQueueHeadTCreate(void)
{
	OS_WaitQueueHeadT *wq = (OS_WaitQueueHeadT *)OS_AllocZero(sizeof(OS_WaitQueueHeadT));
	if(!wq)
		return NULL;
	
	init_waitqueue_head(&wq->t);
	return wq;
}

void OS_WaitQueueHeadTDestroy(OS_WaitQueueHeadT * wq)
{
	if(wq)
		OS_Free(wq);
}

void OS_InitWaitQueueHead(OS_WaitQueueHeadT *wq)	
{
	init_waitqueue_head(&wq->t);
}			
			
int OS_WaitEventInterruptible(OS_WaitQueueHeadT * wq, unsigned int condition)
{
	int ret = 0; 
	if (!(condition)) {
		DEFINE_WAIT(__wait);
		
		for (;;) { 
			prepare_to_wait(&wq->t, &__wait, TASK_INTERRUPTIBLE);
			if (condition)
				break;
				
			if (!signal_pending(current)) { 
				schedule();
				continue;
			} 
			
			ret = -ERESTARTSYS;
			break;
		}
		finish_wait(&wq->t, &__wait);
	}

	return ret;
}

int OS_WaitEventInterruptibleTimeout(OS_WaitQueueHeadT * wq, unsigned int condition, int jtime)
{
	long ret = jtime; 
	
	if(!(condition)) {
		DEFINE_WAIT(__wait);  
		
		for (;;) {
			prepare_to_wait(&wq->t, &__wait, TASK_INTERRUPTIBLE);
			if (condition)
				break;
			if (!signal_pending(current)) {
				ret = schedule_timeout(ret);
				if (!ret)
					break;
				continue;
			}
			ret = -ERESTARTSYS;
			break;
		}
		finish_wait(&wq->t, &__wait);  
	}
	
	return ret;
}

void OS_WakeUpInterruptible(OS_WaitQueueHeadT *wq)
{
	wake_up_interruptible(&wq->t);
}

////////////////////////////////////////////////////////////////////////////////
// Debug
void OS_DebugPrint(const char * pszFormat, ...)
{
	 va_list args;

	 va_start(args, pszFormat);
	 vprintk(pszFormat, args);
	 va_end(args);
}

void OS_BreakOn(bool bBreak, const char * pszFormat, ...)
{
	va_list args;

	if (!bBreak)
		return;

	va_start(args, pszFormat);
	vprintk(pszFormat, args);
	va_end(args);

	BUG();
}

long OS_ISERR(const void *ptr)
{
	return	IS_ERR(ptr);
}

long OS_PTRERR(const void *ptr)
{
	return PTR_ERR(ptr);
}
////////////////////////////////////////////////////////////////////////////////
// Memory
void * OS_Alloc(uint32_t cbAlloc)
{
	return vmalloc_32(cbAlloc);
}

void * OS_AllocZero(uint32_t cbAlloc)
{
	void *pret;
	pret = vmalloc_32(cbAlloc);
	if(NULL == pret)
		return NULL;
		
	memset(pret,0x0,cbAlloc);
	return pret;	
}

void OS_Free(void * pvMem)
{
	vfree(pvMem);
}

void * OS_AlignedAlloc(uint32_t cbSize, uint32_t cbAlign)
{	
	uint8_t *pbRet, *pbAlloc;
	uint64_t qwAddr;
	uint32_t cbExtra = sizeof(uint8_t *) + cbAlign - 1;
	
	pbAlloc = (uint8_t *)OS_Alloc(cbSize + cbExtra);
	if(!pbAlloc)
		return NULL;
	
	qwAddr = (uint64_t)pbAlloc;
	qwAddr = (qwAddr + cbExtra) / cbAlign * cbAlign;
	pbRet = (uint8_t *)qwAddr;	
	*(uint8_t **)(pbRet - sizeof(uint8_t *)) = pbAlloc;
	
	return pbRet;
}

void OS_AlignedFree(void * pvMem)
{
	uint8_t * pbAlloc;
	if(pvMem){
		pbAlloc = *(uint8_t **)((uint8_t *)pvMem - sizeof(uint8_t *));
		OS_Free(pbAlloc);
	}
}

void OS_CopyMemory(void * pvDest, const void * pvSrc, uint32_t cbCopy)
{
	memcpy(pvDest, pvSrc, cbCopy);
}

void OS_MoveMemory(void * pvDest, const void * pvSrc, uint32_t cbCopy)
{
	memmove(pvDest, pvSrc, cbCopy);
}

int OS_CompareMemory(const void * pvMem1, const void * pvMem2, uint32_t cbCompare)
{
	return memcmp(pvMem1, pvMem2, cbCompare);
}

void OS_InitMemory(void * pvDest, uint8_t byData, uint32_t cbInit)
{
	memset(pvDest, byData, cbInit);
}

////////////////////////////////////////////////////////////////////////////////
// Float pointer context protector
void OS_KernelFPUBegin(void)
{
	kernel_fpu_begin();
}

void OS_KernelFPUEnd(void)
{
	kernel_fpu_end();
}

////////////////////////////////////////////////////////////////////////////////
// Delay
void OS_Sleep(int nDurationMS)
{
	msleep(nDurationMS);
}

void OS_SleepUS(int nDurationUS)
{
	udelay(nDurationUS);
}
////////////////////////////////////////////////////////////////////////////////
// I/O
uint32_t OS_ReadReg32(void * pvAddr)
{
	return ioread32(pvAddr);
}

void OS_WriteReg32(void * pvAddr, uint32_t dwData)
{
	iowrite32(dwData, pvAddr);
}

////////////////////////////////////////////////////////////////////////////////
// Spin Lock
struct _OS_SPINLOCK {
	spinlock_t lock;
};

OS_SPINLOCK * OS_SpinLockCreate(void)
{
	OS_SPINLOCK * pSpinLock = (OS_SPINLOCK *)OS_AllocZero(sizeof(OS_SPINLOCK));
	if(!pSpinLock)
		return NULL;
		
	spin_lock_init(&pSpinLock->lock);
	return pSpinLock;
}

void OS_SpinLockDestroy(OS_SPINLOCK * pSpinLock)
{
	OS_Free(pSpinLock);
}

void OS_SpinLock(OS_SPINLOCK * pSpinLock)
{
	spin_lock(&pSpinLock->lock);
}

void OS_SpinUnlock(OS_SPINLOCK * pSpinLock)
{
	spin_unlock(&pSpinLock->lock);
}

////////////////////////////////////////////////////////////////////////////////
// Thread
struct _OS_Thread {
	struct task_struct *pthread;
};

OS_Thread* OS_CreateThread(int (*threadfn)(void *data),
				   void *data,
				   const char namefmt[])
{
	OS_Thread * p = (OS_Thread *)OS_AllocZero(sizeof(OS_Thread));
	if(!p)
		return NULL;
		
	p->pthread = kthread_run(threadfn, data, namefmt);
	if(!p->pthread) {
		OS_Free(p);
		return NULL;
	}
	
	return p;
}

int OS_StopThread(OS_Thread * p)
{
	int ret;
	if(p && p->pthread){
		ret = kthread_stop(p->pthread);
		OS_Sleep(10);
		OS_Free(p);
		return ret;
	}
	
	return -EINVAL;	
}

bool OS_IsThreadShouldStop(void)
{
	return kthread_should_stop();
}

void *OS_DMABufferCreate(void *hwdev, size_t size,
		     dma_addr_t *dma_handle)
{
	return pci_alloc_consistent((struct pci_dev *)hwdev, size, dma_handle);
}

void OS_DMABufferDestroy(void *hwdev, size_t size,
		    void *vaddr, dma_addr_t dma_handle)
{
	pci_free_consistent((struct pci_dev *)hwdev, size, vaddr, dma_handle);
}

void OS_SetFreezable(void)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22)
	set_freezable();
#endif	
}

void OS_TryToFreeze(void)
{
	might_sleep();
	if(likely(!freezing(current)))
		return ;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22)
	try_to_freeze();
#endif	
}

////////////////////////////////////////////////////////////////////////////////
//  div64
uint64_t OS_Div64(uint64_t dividend, uint64_t divisor)
{
	do_div(dividend,divisor);
	return dividend;
}

////////////////////////////////////////////////////////////////////////////////
// Event 
struct _OS_Completion {
	struct completion c;
};

OS_Completion * OS_CreateEvent(void)
{
	OS_Completion *pCompletion = (OS_Completion *)OS_Alloc(sizeof(OS_Completion));
    if (NULL == pCompletion)
        return NULL;

    init_completion(&pCompletion->c);
    return pCompletion;	
}

void OS_DestroyEvent(OS_Completion * pEvent)
{
    OS_Free((OS_Completion *)pEvent);
}

int OS_WaitForEvent(OS_Completion * pEvent, uint32_t dwTimeoutJiffies)
{
	struct completion * pc = &pEvent->c; 
    return wait_for_completion_timeout(pc, dwTimeoutJiffies);
}

void OS_SetEvent(OS_Completion * pEvent)
{
	struct completion * pc = &pEvent->c; 
    complete(pc);
}

void OS_ResetEvent(OS_Completion * pEvent)
{
	struct completion * pc = &pEvent->c; 
	pc->done = 0;
}
